checkMendelian<-function(geneSetObj, quiet=TRUE)
{
  if(!quiet)
    cat("converting geneSet object to numerical matrix...\n")
  pedObj<-geneSet2Ped(geneSetObj)

  if(!quiet)
    cat("checking Mendelian errors ...\n")

  res<-checkMendelian.default(pedObj, quiet)
  return(res)
}

checkMendelian.default<-function(pedObj, quiet=TRUE)
{
  pedObj <- checkColumn(pedObj)
  ped<-pedObj$ped
  columns<-pedObj$columns

  # check if number of alleles of a marker > 2
  tmp<-getAlleles(pedObj)
  alleles<-tmp$alleles
  nAllele<-tmp$nAllele
  family<-as.numeric(ped[,1])
  family.u<-unique(family) # unique family id
  nFamily<-length(family.u) # number of family
  familySize<-tapply(rep(1,length(family)), family, sum) # number of members of families

  if(!quiet)
    cat("checking compatibility ...\n")

  # check compatibility
  sampleInfoLen<-length(columns)
  res<-checkCompatibility(ped, alleles, nAllele, sampleInfoLen, nFamily, 
                          familySize, pedObj$markerNames, quiet)
  return(res)
}

checkCompatibility <- function(ped, alleles, nAllele, sampleInfoLen, 
                               nFamily, familySize, markerNames, 
                               quiet=FALSE)
{
  nr <- nrow(ped)
  nc <- ncol(ped)

  ## sort ped according to pedigree id (i.e. family)
  ped2 <- ped[order(ped[, 1]), ,drop=FALSE]

  nr<-as.integer(nr)
  nc<-as.integer(nc)
  ped2<-as.integer(as.vector(t(ped2)))
  alleles<-as.integer(alleles)
  nAllele<-as.integer(nAllele)
  nFamily<-as.integer(nFamily)
  familySize<-as.integer(familySize)
  sampleInfoLen<-as.integer(sampleInfoLen)
  nMarkers<- as.integer(nc-sampleInfoLen)/2

  res<-.C("checkCompatible",
          ped2,
          nr,
          nc,
          nFamily,
          familySize,
          alleles,
          nAllele,
          sampleInfoLen,
          compatibleFlag=as.integer(1),
          errorFlag=as.integer(0),
          whichMarker=as.integer(0),
          nMerrMarker=as.integer(rep(0, nMarkers)),
          nMerrFamily=as.integer(rep(0, nFamily)),
          nErrFamilySample=as.integer(rep(0, nFamily)),
          quiet=as.integer(quiet),
          PACKAGE="fbat"
          )

  compatibleFlag <- res$compatibleFlag
  errorFlag <- res$errorFlag

  if(compatibleFlag==0 && !quiet)
      warning(
              "One or more sets of parent genotypes are not compatible ", 
              "with the corresponding child's genotypes. Such ", 
              "nuclear families will be excluded from analysis."
              )

  nMerrMarker<-res$nMerrMarker
  names(nMerrMarker)<-markerNames
  nMerrFamily<-res$nMerrFamily
  names(nMerrFamily)<-paste("family", 1:nFamily, sep="")
  nErrFamilySample<-res$nErrFamilySample
  names(nErrFamilySample)<-paste("family", 1:nFamily, sep="")

  res<-list(errorFlag=errorFlag, 
            compatibleFlag=compatibleFlag,
            nMerrMarker=nMerrMarker,
            nMerrFamily=nMerrFamily,
            nErrFamilySample=nErrFamilySample)
  return(res)
}

