## read .ped file
readPed <- function(
                          filename, 
                          columns=c("family", "pid", "father", "mother", 
                                    "sex", "affected"), 
                          quiet=FALSE 
                          )
{
  ## pedigree file format
  ## --------------------
  ##
  ## line 1         : list of genetic markers
  ## line 2 to (N+1): <family> <pid> <father> <mother> <sex> [<affected>] 
  ##                  <genotype_1> ... <genotype_n>
  ##
  ## some pedigree files contain <affected> after <sex> and before
  ## <genotype_1>, some pedigree files do not.
  ##
  ## <family> is a unique identifier for each family, and within each family
  ## <pid> is a unique identifier for an individual.
  ## <father> and <mother> identify the individuals father and mother
  ##    (if this line refers to a founder, these should be set to
  ##     zero).
  ##  <sex> denotes the individuals sex, using the convention
  ##     1=male, 2=female.
  ##
  ## Each <genotype> is encoded as two integer allele numbers.

  # example file (from R GeneticsBase package)
  # http://www.bioconductor.org/packages/2.0/bioc/html/GeneticsBase.html

  #m709 m654 m47 p46 p79 p252 p491 p523
  #1 1900 0 0 1 0 1 1 1 1 1 1 2 2 2 2 2 2 1 1 1 1
  #1 1667 1900 978 2 2 1 1 1 1 1 2 2 2 2 1 1 2 1 1 1 2
  #1 978 0 0 2 0 1 1 1 1 1 2 2 2 2 1 1 2 1 1 1 2
  #2 1391 0 0 1 0 1 1 1 1 1 2 2 2 2 1 1 2 1 1 1 2
  #2 1988 0 0 2 0 1 1 2 1 2 2 1 2 1 1 1 2 1 1 1 2
  #2 109 1391 1988 2 2 1 1 1 1 1 2 2 2 2 1 0 0 1 1 1 2
  
  filename=as.character(filename)

  len <- length(columns)

  ## read numer of rows, columns and markers
  res <- .C(
            "getDimOfFile", 
            filename=filename, 
            nrow=as.integer(0), 
            sampleInfoLen=as.integer(len), 
            nMarkers=as.integer(0), 
            errorFlag=as.integer(0),
            PACKAGE="fbat"             
            )

  errorFlag=res$errorFlag
  if(errorFlag!=0)
    stop("Unable to determine file dimensions for ped file `", filename, "'.")

  nrow=res$nrow
  if(nrow==0)
    stop("No data in ped file `", filename, "'.")

  nMarkers=res$nMarkers
  if(nMarkers==0)
    stop("No marker information in ped file `", filename, "'.")

  ncol <- 2*nMarkers+length(columns)

  ped <- rep(0, nrow*ncol)
  strSpace <- paste(rep(" ", 60), collapse='')
  markerNames <- rep(strSpace, nMarkers)

  res.getPed <- .C(
                   "getPed", 
                   filename=filename, 
                   ped=as.integer(ped), 
                   nrow=as.integer(nrow), 
                   ncol=as.integer(ncol), 
                   nMarkers=as.integer(nMarkers), 
                   markerNames=as.character(markerNames), 
                   errorFlag=as.integer(0),
                   PACKAGE="fbat"             
                   )

  errorFlag=res.getPed$errorFlag
  if(errorFlag!=0)
    stop("Unable to load data from `", filename, "'")

  ped <- res.getPed$ped
  ped <- matrix(ped, nrow=nrow, byrow=T)
  markerNames <- res.getPed$markerNames
  nMarkers <- length(markerNames)

  markercols <- c(
                  sapply(markerNames, 
                         function(x) paste(x, c("1", "2"), sep=".")
                         )
                  )
  ## Add these to the other columns
  colNames <- c(columns, markercols)

  ped <- matrix(ped, nrow=nrow, ncol=ncol) 
  colnames(ped) <- colNames
  rownames(ped) <- paste(ped[, 1,drop=FALSE], ped[, 2,drop=FALSE], sep=".")

  ## we can do checking in analysis
  #tmp <- checkColumn(ped, columns)


  ## obtain unique nuclear family id
  family <- as.vector(ped[, 1,drop=FALSE])
  family.u <- unique(family)

  ## obtain sizes (number of members) of nuclear families
  familySize <- tapply(family, family, length)

  ## obtain the number of nuclear families
  nFamily <- length(family.u)

  if(!quiet)
  { 
    cat("Successfully read the pedigree file `", filename, "'.\n");
    cat("\n")
    cat("Number of Markers: ", length(markerNames), "\n")
    cat("Number of Subjects:", nrow(ped), "\n")
    cat("Number of Families:", nFamily, "\n")
    cat("\n")
  }

  Position<-rep("?", length(markerNames))
  pedObj<-list(ped=ped, columns=columns, markerNames=markerNames, Position=Position, fileName=filename)
  
  invisible(pedObj)
}

