/**
 * C functions used for Family Based Association Test.
 * Created by Weiliang Qiu on March 8, 2005
**/

#include "FBAT.h"

/*
 * Given n nuclear families, do FBAT test.
 *
 * ped -- nrow by ncol matrix of n nuclear families
 * nrow -- total number of individuals in these n nuclear families
 * ncol -- 2*nMarker+6, where nMarker is the number of markers. 
 * nFamily -- number of nuclear families
 * familySize -- nFamily by 1 vector of sizes of nuclear families. Size of a nuclear
 *      family is the number of individuals in this family.
 * genoMethod -- genotype coding method
 * traitMethod -- trait coding method
 * traitOffset -- offset for the trait
 * alleles -- vector of alleles for markers. alleles of a marker are followed
 *            by alleles of another marker.
 * nAllele -- nMarker by 1 vector. number of alleles for each marker.
 *
 * Outputs:
 * S -- a M by nX matrix of scores. each row corresponds to a marker.
 *       S[M,]=sum_{j}^{nChildrenM} X[j]*T[j], 
 *   where X[j] and T[j] are the observed coded genotype and trait, 
 *   respectively, for the j-th children. 
 *   'nX'='nAllelesM' or 'nAllelesM*(nAllelesM+1)/2'
 * ES -- a M by nX matrix of conditional expectations of S[1,],...,S[M,]
 *       'nX'='nAllelesM' or 'nAllelesM*(nAllelesM+1)/2' 
 * CovS -- a M by nX by nX array of conditional covariance of S[1,],...,S[M,]
 *       'nX'='nAllelesM' or 'allelesM*(nAllelesM+1)/2' 
 */
void fbat(int *ped, int *nrow, int *ncol, int *nFamily, int *familySize, 
		int *genoMethod, int *traitMethod, double *traitOffset, 
		int *alleles, int *nAllele, double *S, double *ES, 
		double *CovS, int *numInfoFamily, 
		int *errorFlag, int *quiet)
{
  int i, j, M, nnFamily, nnrow, nncol, nMarker;
  int *pedM, start, end, nnAllele, nX, nX2, nXsum, nXsum2; 
  int *tmpAlleles, start2, end2, numInfoM;
  double zero, *SM, *ESM, *CovSM;
  
  nnFamily=*nFamily; // number of nuclear families
  nnrow=*nrow; // number of individuals in the pedigree file
  nncol=*ncol; // number of columns of the pedigree file
  nMarker=(nncol-6)/2; // number of markers
  zero=0.0;
  *errorFlag=0;
  start2=0;

  //input first 6 columns of pedigrees
  //For each marker, the pedigree contains 8 columns
  //First 6 columns are 'pid', 'id', 'fid', 'mid', 'sex', 'aff'.
  //The rest two columns are two alleles of the marker
  pedM=(int *)gS_alloc((size_t)nnrow*8, sizeof(int));
  for(i=0; i<nnrow; i++)
  { for(j=0; j<6; j++)
    { 
      pedM[i*8+j]=ped[i*nncol+j]; 
    }
    //fid=id or mid=id
    if(pedM[i*8+2]==pedM[i*8+1])
    {
      if(*quiet==0)
      { warning("Warning! father id = subject id = %d!", pedM[i*8+1]); }
      *errorFlag=3;
      //free(pedM); // Automatically performed for 'S_alloc'
      return;
    }
    if(pedM[i*8+3]==pedM[i*8+1])
    {
      if(*quiet==0)
      { warning("Warning! mother id = subject id = %d!", pedM[i*8+1]); }
      *errorFlag=3;
      //free(pedM); // Automatically performed for 'S_alloc'
      return;
    }
  }

  nXsum=0;
  nXsum2=0;
  //for each marker do FBAT analysis
  for(M=0; M<nMarker; M++)
  {
    //input the genotype info
    start=6+M*2;
    end=6+M*2+1;
    for(i=0; i<nnrow; i++)
    { pedM[i*8+6]=ped[i*nncol+start]; 
      pedM[i*8+7]=ped[i*nncol+end]; 
    }

    //create temporary score, expected score and covariance matrix
    //Note that for different markers, the number of alleles may be different
    //'allele' is a vector consecutively recording the alleles of all markers 

    nnAllele=nAllele[M]; // number of alleles for this marker
    if(nnAllele>2){
      if(*quiet==0)
      { warning("Warning! Marker %d has more than 2 alleles!", (M+1));
        warning("Current version only supports bi-allelic markers!");
        warning("Program terminated!");
      }
      //free(pedM); // Automatically performed for 'S_alloc'
      *errorFlag=2;
      return;
    }
    
    //get the alleles for this marker
    tmpAlleles=(int *)gS_alloc((size_t) nnAllele, sizeof(int));
    if(M>0)
    { end2=start2+nAllele[M]; 
    } else {
      start2=0;
      end2=nAllele[M];
    }
    j=0;
    for(i=start2; i<end2; i++)
    { tmpAlleles[j]=alleles[i]; 
      j++; 
    }
    
    start2=end2;

    //'nX' is the length of genotype score vector
    if(*genoMethod!=3)
    { nX=nnAllele; }
    else { nX=nnAllele*(nnAllele+1)/2; } // GEN genotype coding

    nX2=nX*nX;
    // score vector for the marker
    SM=(double *)gS_alloc((size_t) nX, sizeof(double));
    // expected score vector for the marker
    ESM=(double *)gS_alloc((size_t) nX, sizeof(double));
    // covariance matrix of the score vector for the marker
    CovSM=(double *)gS_alloc((size_t) nX2, sizeof(double));
    
    //do FBAT analysis
    fbatM(pedM, &M, nFamily, familySize, genoMethod, &nX, traitMethod, 
		    traitOffset, tmpAlleles, &nnAllele, SM, ESM, CovSM, 
		    &numInfoM, quiet); 
    //number of informative family for maker M
    numInfoFamily[M] = numInfoM;

    // record scores, expected scores and covariance matrix
    for(i=0; i<nX; i++)
    {
      S[nXsum+i]=SM[i];
      ES[nXsum+i]=ESM[i];
    }
    for(i=0; i<nX2; i++)
    {
      CovS[nXsum2+i]=CovSM[i];
    }
    nXsum+=nX;
    nXsum2+=nX2;


    //free(SM); free(ESM); free(CovSM); free(tmpAlleles); // Automatically performed for 'S_alloc'
  }
  //free(pedM); // Automatically performed for 'S_alloc'

  return;
}




/*
 * Given a nuclear family with complete parental genotypes,
 * calculate the conditional probability of child j's genotypes conditional
 * on parental genotypes under null hypothesis.
 *
 * Alleles are represented by positive integers. Use zero for missing alleles.
 * We assume that data are 'clean', i.e., all genotypes are compatible
 * and there is no missing alleles.
 *
 * parents -- 2 by 2 matrix, row 1 is the two alleles of father
 *           row 2 is the two alleles of mother;
 * childj -- 1 by 2 matrix, the two alleles of child j;
 * prob -- the conditional probability of child j's genotype.
 *
 * 'condProbCPj' represents {cond}itional {Pro}bability with {C}omplete
 * {P}arental genotypes for child {j}.
*/
void condProbCPj(int *parents, int *childj, double *prob, int *compatibleFlag)
{
  int flag;

  // check compatibility of genotypes of parents and a child
  checkCompatibility1(parents, childj, &flag);
  if(flag==0) // not compatible
  { *compatibleFlag=0;
    *prob=0.0;
    return;
  }
  *compatibleFlag=1;
  if((parents[0*2+0]!=parents[0*2+1]) && (parents[1*2+0]!=parents[1*2+1]))
  {  /* parents' genotype are both heterozygous */
    if((parents[0*2+0]+parents[0*2+1])==(parents[1*2+0]+parents[1*2+1]))
    { /* father's genotype is equal to mother's genotype */
      if(childj[0]!=childj[1]) /* child j is heterozygous */ 
      { *prob=0.5;
        return; 
      } 
      *prob=0.25; /* child j is homozygous */
      return;
    }

    /* father's genotype is not equal to mother's genotype */
    *prob=0.25;
    return;
  }
  if((parents[0*2+0]==parents[0*2+1]) && (parents[1*2+0]==parents[1*2+1]))
  { /* parents' genotype are both homozygous */
    *prob=1.0;
    return;
  }
  /* one parent is homozygous and the other is heterozygous */
  *prob=0.5;
  return;
}

/*
 * Given a nuclear family with complete parental genotypes,
 * check if parental genotypes and a child genotypes are compatible.
 * We assume that both parents genotype are complete
 *
 * Alleles are represented by positive integers. Use zero for missing alleles.
 *
 * parents -- 2 by 2 matrix, row 1 is the two alleles of father
 *           row 2 is the two alleles of mother;
 * childj -- 1 by 2 matrix, the two alleles of child j;
 * flag -- flag=0 indicates compatible; flag!=0 indicates incompatible
 *
*/
void checkCompatibility1(int *parents, int *childj, int *flag)
{
  int p1, p2, q1, q2, p1q1[2], p1q2[2], p2q1[2], p2q2[2];
  int c1, c2;

  p1=parents[0*2+0];
  p2=parents[0*2+1];
  q1=parents[1*2+0];
  q2=parents[1*2+1];

  p1q1[0]=minI(p1, q1); p1q1[1]=maxI(p1, q1);
  p1q2[0]=minI(p1, q2); p1q2[1]=maxI(p1, q2);
  p2q1[0]=minI(p2, q1); p2q1[1]=maxI(p2, q1);
  p2q2[0]=minI(p2, q2); p2q2[1]=maxI(p2, q2);

  c1=minI(childj[0], childj[1]); 
  c2=maxI(childj[0], childj[1]);

  if( (c1==p1q1[0] && c2==p1q1[1]) ||
      (c1==p1q2[0] && c2==p1q2[1]) ||
      (c1==p2q1[0] && c2==p2q1[1]) ||
      (c1==p2q2[0] && c2==p2q2[1]) )
  { *flag=1; } // compatible
  else { *flag=0; }
  return;
}

/*
 * Given a nuclear family with one parental genotype missing,
 * check if parental genotypes and a child genotypes are compatible.
 * We assume that both parents genotype are complete
 *
 * Alleles are represented by positive integers. Use zero for missing alleles.
 *
 * parents -- 1 by 2 matrix. 
 * childj -- 1 by 2 matrix, the two alleles of child j;
 * flag -- flag=0 indicates compatible; flag!=0 indicates incompatible
 *
*/
void checkCompatibility2(int *parents, int *childj, int *flag)
{
  int flag1, flag2;

  flag1=0;
  if(childj[0]==parents[0] || childj[0]==parents[1])
  { flag1=1; }
  flag2=0;
  if(childj[1]==parents[0] || childj[1]==parents[1])
  { flag2=1; }
  *flag=flag1+flag2;
  
  return;
}

/* count the number of different alleles of children in a nuclear family 
 */
int *countAlleles(int *children, int *nChildren, int *nAllele)
{
  int i, j, nnChildren, n, m, nnAllele, *alleles, flag2;

  nnChildren=*nChildren;
  m=nnChildren*2;
  alleles=(int *)gS_alloc((size_t)m, sizeof(int));

  if(children[0*2+0]!=children[0*2+1])
  { nnAllele=2;
    alleles[0]=children[0*2+0];
    alleles[1]=children[0*2+1];
  } else {
    nnAllele=1;
    alleles[0]=children[0*2+0];
  }
  n=nnChildren*2;
  for(i=2; i<n; i++)
  { 
    flag2=0;
    for(j=0; j<nnAllele; j++)
    { if(children[i]==alleles[j])
      { continue; }
      else { flag2++; }
    }
    if(flag2==nnAllele) // children[i] is a new allele
    { 
      alleles[nnAllele]=children[i];
      nnAllele++; 
    }
  }
  *nAllele=nnAllele;
  if(nnAllele<m) {
    alleles=(int *)gS_realloc(alleles, 
                             (size_t) (nnAllele),
                             (size_t) m,
                             (size_t) sizeof(int));
  }
	
  return(alleles);
}

/*
 * Given a nuclear family with one missing parental genotype and one 
 * homozygous parental genotype, calculate the joint conditional probability of 
 * child j and child k's genotypes under null hypothesis conditional on 
 * one parent's genotype is missing, the other parent's genotype is 
 * homozygous, and the configuration of offspring's genotype is given.
 *
 * Alleles are represented by positive integers. Use zero for missing alleles.
 * We assume that data are 'clean', i.e., all genotypes are compatible
 * and there is no missing alleles.
 *
 * n -- number of observed offsprings in this nuclear family;
 * m -- number of elements in the observed offspring genotype configuration,
 *      'm' can take only values 1 or 2; That is, 'm' is the number of 
 *      different genotypes for the observed offsprings
 * childj -- 1 by 2 matrix, the two alleles of child j;
 * childk -- 1 by 2 matrix, the two alleles of child k;
 * probj -- the conditional probability of child j's genotype;
 * probk -- the conditional probability of child k's genotype;
 * probjk -- the joint conditional probability of children j and k's genotypes.
*/
void condProbM1PHomojk(int *n, int *m, int *childj, int *childk, 
		double *probj, double *probk, double *probjk)
{
  int nn, flag;
  double numer, denom, tmp;

  nn=*n;

  /* the offspring genotype configuration contains only one genotype */
  /* For example, if the parent genotype is AA, then compatible
   * offspring genotype configurations which contains only one genotype are
   * {AA}, {AB}, or {AC}
   */
  if(*m==1)
  { flag=checkEqualGeno(childj, childk);
    if(flag==0)
    { printf("Error! two children's genotypes are not compatible!\n");
      printf("Program terminated!\n");
      *probj= -1.0;
      *probk= -1.0;
      *probjk= -1.0;
      return;
    }
    *probj=1.0;
    *probk=1.0;
    *probjk=1.0;
    return;
  }
  /* the offspring genotype configuration contains two genotypes */
  /* For example, if the parent genotype is AA, then compatible
   * offspring genotype configurations which contains two genotypes are
   * {AA, AB}, {AA, AC}, {AB, AC}
   */
  *probj=0.5;
  *probk=0.5;
  //if nn is large, then 2^(nn-2) will be huge.
  //So we re-express the formula.
  tmp=1.0/pow(2.0, nn-2.0);
  denom=2.0*(2.0-tmp);
  flag=checkEqualGeno(childj, childk);
  if(flag==1)
  { /* children j and k have the same genotype, e.g. {AC, AC}*/
    numer=(1.0-tmp);
    //*probjk=(pow(2.0, nn-2.0) - 1.0) / (pow(2.0, nn) - 2.0);
  } else { //e.g. {AB, AC}
    numer=1.0;
    //*probjk=(pow(2.0, nn-2.0)) / (pow(2.0, nn) - 2.0);
  }
  *probjk=numer/denom;
  return;
}

int checkEqualGeno(int *Geno1, int *Geno2)
{
  int p1, p2, q1, q2;

  p1=Geno1[0]; p2=Geno1[1];
  if(p1>p2)
  { p1=Geno1[1]; p2=Geno1[0]; }

  q1=Geno2[0]; q2=Geno2[1];
  if(q1>q2)
  { q1=Geno2[1]; q2=Geno2[0]; }

  if(p1==q1 && p2==q2)
  { return(1); }
  else { return(0); }
}


/*
 * Given a nuclear family with one missing parental genotype and one 
 * heterozygous genotype (or with no parental genotype available),
 * calculate the joint conditional probability of 
 * child j and child k's genotypes under null hypothesis conditional on 
 * one parent's genotype is missing, the other parent's genotype is 
 * heterozygous (or conditional on no parental genotype is availabe), 
 * and the configuration of offspring's genotype is given.
 *
 * Currently, we only consider biallelic cases.
 *
 * Alleles are represented by positive integers. Use zero for missing alleles.
 * We assume that data are 'clean', i.e., all genotypes are compatible
 * and there is no missing alleles.
 *
 * n -- number of observed offsprings in this nuclear family;
 * m -- number of elements in the observed offspring genotype configuration,
 *      'm' can take only values 1 or 2;
 * config -- m by 2 matrix of the observed offspring genotype configuration, 
 *           row k is the two alleles of the k-th element in the 
 *           offspring genotype configuration;
 * nAA -- number of homozygous offsprings in this nuclear family
 * nAB -- number of heterozygous offsprings in this nuclear family
 * childj -- 1 by 2 matrix, the two alleles of child j;
 * childk -- 1 by 2 matrix, the two alleles of child k;
 * probj -- the conditional probability of child j's genotype;
 * probk -- the conditional probability of child k's genotype;
 * probjk -- the joint conditional probability of children j and k's genotypes.
*/
void condProbMPHeterojk(int *n, int *m, int *config, int *nAA, 
		int *nAB, int *childj, int *childk, double *probj, 
		double *probk, double *probjk)
{
  double numer, denom;
  int mm, nn, nnAA, nnAB, flag;

  mm=*m;
  nn=*n;
  nnAA=*nAA;
  nnAB=*nAB;

  if(mm==1)
  { 
    flag=checkEqualGeno(childj, childk);
    if(flag==0)
    { printf("Error! two children's genotypes are not compatible!\n");
      printf("Program terminated!\n");
      *probj = -1.0;
      *probk = -1.0;
      *probjk = -1.0;
      return;
    }	  
    /* the offspring genotype configuration contains only one genotype */
    /* for example, parent's genotype is AB, the offspring genotype 
     * configuration is {AA} or {AB} or {BB}
     */
	  
    *probj=1.0;
    *probk=1.0;
    *probjk=1.0;
    return;
  }
  /* the offspring genotype configuration contains two genotypes */
  /* for example, parent's genotype is AB, the offspring genotype 
   * configuration is {AA, AB}, {AA, BB}, {AB, BB}
   */
  //denom=pow(4.0, nn) - 2 * pow(3.0, nn) + pow(2.0, nn);
  denom=1.0-2.0*pow(0.75, nn)+pow(0.5, nn);
  
  if( ((config[0*2+0]==config[0*2+1]) && 
			  (config[1*2+0]==config[1*2+1])) || mm==3)
  { /* all genotypes in the offspring genotype configuration are homozygous */
    /* for example: {AA, BB} */
    /* Or the offspring genotype configuration contains all possible genotypes
     * For example: {AA, AB, BB}
     */

    if(childj[0]==childj[1])
    { numer=(1-pow(0.75, nn-1.0))/4.0; }
    else {
      numer=0.5-pow(0.75, nn-1.0)+pow(0.5, nn);
    }
    *probj=numer/denom;
    if(childk[0]==childk[1])
    { numer=(1-pow(0.75, nn-1.0))/4.0; }
    else {
      numer=0.5-pow(0.75, nn-1.0)+pow(0.5, nn);
    }
    *probk=numer/denom;

    //calculate probjk
    if((childj[0]==childj[1]) && (childk[0]==childk[1]))
    { /* children j and k are both homozygous */
      if(childj[0]==childk[0])
      { /* children j and k have the same homozygous genotype */
        /* for example, {AA, AA} */
        //numer=pow(4.0, nn-2.0) - pow(3.0, nn-2.0);
	numer=(1-pow(0.75, nn-2.0))/16.0;
      } else { /* children j and k have different homozygous genotype */
               /* for example, {AA, BB} */
        //numer=pow(4.0, nn-2.0);
	numer=1/16.0;
      }
    } else if ((childj[0]!=childj[1]) && (childk[0]!=childk[1])) 
    { /* children j and k are both heterozygous */
      /* for example, {AB, AB} */
      //numer=pow(4.0, nn-1.0)-8*pow(3.0, nn-2.0)+pow(2.0, nn);
      numer=0.25-0.5*pow(0.75, nn-2.0)+pow(0.5, nn);
    } else { /* one child is homozygous and one child is heterozygous */
	    /* for example, {AA, AB} */
      //numer=pow(4.0, nn-2.0) - pow(3.0, nn-2.0);
      numer=(1-pow(0.75, nn-2.0))/8.0;
    }
    *probjk=numer/denom;
    return;
  }
  /* The offspring genotype configuration contains at leat one homozygous 
   * genotype and at least one heterozygous genotype
   * For example: {AA, AB}
   */
  if(childj[0]==childj[1])
  { *probj=nnAA*1.0/nn; }
  else {
    *probj=nnAB*1.0/nn;
  }
  if(childk[0]==childk[1])
  { *probk=nnAA*1.0/nn; }
  else {
    *probk=nnAB*1.0/nn;
  }
  
  if((childj[0]==childj[1]) && (childk[0]==childk[1]))
  { /* children j and k are both homozygous */
    //*probjk=nnAA*(nnAA-1.0)/(nn*(nn-1.0));
    *probjk=(nnAA*1.0/nn)*((nnAA-1.0)/(nn-1.0));
  } else if((childj[0]!=childj[1]) && (childk[0]!=childk[1]))
  { /* children j and k are both heterozygous */
    //*probjk=nnAB*(nnAB-1.0)/(nn*(nn-1.0));
    *probjk=(nnAB*1.0/nn)*((nnAB-1.0)/(nn-1.0));
  } else {
    //*probjk=nnAA*nnAB*1.0/(nn*(nn-1.0));
    *probjk=(nnAA*1.0/nn)*(nnAB*1.0/(nn-1.0));
  }
  return;
}

/*
 * transform the trait Y_{ij} of the j-th individual in the i-th family
 *
 * Y -- trait value (2 (affected), 1 (unaffected), 0 (unknown))
 * offset -- offset for the trait value
 * method -- coding method
 * T -- code of the trait
 */
void Tfun(double *Y, double *offset, int *method, double *T)
{
  switch(*method){
    case 1: 
	    *T=*Y-*offset;
	    break;

    default: //dichotomous trait
	    if((int)(*Y)==2) // affected
	    { *T=1.0; }
	    else // unaffected or unknown
	    { *T=0.0; }
	    break;
  }
}

/*
 * transform the genotype g_{ij} of the j-th individual in the i-th family
 * at a given marker
 *
 * g -- two alleles of the genotype at a given marker
 * alleles -- set of alleles
 * nAllele -- number of possible alleles
 * method -- coding method; only allow GDOM, GREC, and GTDT 
 *           see Schaid (1996). "General Score Tests for Associations
 *           of Genetic Markers With Disease Using Cases and Their Parents",
 *           Genetic Epidemiology 13:423-449.
 *           Default coding method is GTDT.
 * nX -- number of elements of X
 * X -- code of the genotype
 */
void Xfun(int *g, int *alleles, int *nAllele, int *method, int *nX, double *X)
{
  int j, nnAllele, one, pos, allele1, allele2, nnX, pos1, pos2;
  double zero;

  nnAllele=*nAllele;
  nnX=*nX;
  one=1;
  zero=0.0;

  // initialize X
  matIni(X, &zero, &one, nX);

  switch(*method) {
	case 1: //GDOM: X[j]=1 if g has one or two alleles of type j,
		// Otherwise X[j]=0.
		  for(j=0; j<nnX; j++)
		  { if(g[0]==alleles[j] || g[1]==alleles[j])
		    { X[j]=1; }
		  }
		  break;

	case 2: // GREC: X[j]=1 if g has two alleles of type j,
		// Otherwise X[j]=0.
		  for(j=0; j<nnX; j++)
		  { if(g[0]==alleles[j] && g[1]==alleles[j])
		    { X[j]=1; }
		  }
		  break;

	case 3: // GEN: length of X is equal to the number of genotypes 
		  // i.e., nnAllele*(nnAllele+1)/2. 
		  //find position of g[0] and g[1] in alleles
		  pos1=0; pos2=0;
		  for(j=0; j<nnAllele; j++)
		  { if(g[0]==alleles[j]) { pos1=j+1; }
		    if(g[1]==alleles[j]) { pos2=j+1; }
		  }
                  allele1=pos1;
                  allele2=pos2;
                  /** GEN method requires allele1<=allele2 to facilitate 
		   * calculation **/
                  if(pos1>pos2)
                  { allele1=pos2;
                    allele2=pos1;
                  }

		  pos=nnAllele*(allele1-1)-(allele1-1)*(allele1-2)/2+
			  (allele2-allele1);
		  X[pos]=1.0;
		  break; 

	default: // GTDT: X[j]= # of alleles of type j in the genotype g.
		  for(j=0; j<nnX; j++)
		  { if(g[0]!=alleles[j] && g[1]!=alleles[j])
		    { X[j]=0; }
		    else if(g[0]==alleles[j] && g[1]==alleles[j])
		    { X[j]=2; }
		    else {
		      X[j]=1;
		    }
		  }
		  break;
  }
  return;
}


/*
 * calculate the statistic for the i-th family at a given marker 
 *
 * Inputs:
 * X -- nChildren by ncolX matrix of coded genotypes of nChildren children
 * ncolX -- dimension of the coded genotype
 * T -- nChildren by 1 matrix of coded traits of nChildren children
 * nChildren -- number of children
 * 
 * Outputs:
 * Si -- 1 by ncolX vector of statistic
 * sumTi -- sum_{j}^{nChildren}T[j]
 * sumTi2 -- sum_{j}^{nChildren}T[j]^2
 */
void getSiFun(double *X, int *ncolX, double *T, int *nChildren, double *Si, double *sumTi, double *sumTi2)
{
  int nncolX, nnChildren, j, k;

  nncolX=*ncolX;
  nnChildren=*nChildren;

  //Si=sum_{j}^{nChildren} X[j]*T[j]
  for(k=0; k<nncolX; k++)
  { Si[k]=T[0]*X[0*nncolX+k]; }

  *sumTi=T[0];
  *sumTi2=T[0]*T[0];
  for(j=1; j<nnChildren; j++)
  {
    for(k=0; k<nncolX; k++)
    { Si[k] += T[j]*X[j*nncolX+k]; }
    *sumTi += T[j];
    *sumTi2 += T[j]*T[j];
  }
  return;
}

/*
 * calculate conditional expectation E(Si)
 *
 * Inputs:
 * X -- nP by ncolX matrix of coded genotype of all children that are
 *      possible in the nuclear family. 
 * ncolX is the dimension of a coded genotype.
 * Prob -- conditional probability of child's genotype
 * nP is the number of genotypes that 
 *      are possible in the nuclear family; 
 * sumTi -- sum_{j}^{nChildren}T[j]
 *
 * Outpts:
 * ESi -- conditional expectation of Si.
 *        ESi=(sum_{j}^{nChildren}T[j])*(sum_{g}^{G}X[g]*Prob[g])
 */
void getESiFun(double *X, int *ncolX, double *Prob, int *nP, double *sumTi,
		double *ESi)
{
  int j, g, G, nncolX;
  double sumProb;

  G=*nP;
  nncolX=*ncolX;

  sumProb=0.0;
  for(g=0; g<G; g++)
  { sumProb += Prob[g]; }
  if(fabs(sumProb-1.0)>1.0e-6)
  { error("Sum of all probabilities (%lf) not equal to 1!", sumProb);
  }

  for(j=0; j<nncolX; j++)
  { ESi[j]=X[0*nncolX+j]*Prob[0]; }

  for(g=1; g<G; g++)
  { for(j=0; j<nncolX; j++)
    { ESi[j] += X[g*nncolX+j]*Prob[g]; }
  }
  for(j=0; j<nncolX; j++)
  { ESi[j]=(*sumTi)*ESi[j]; }
  return;
}

/*
 * calculate conditional covariance matrix Cov(Si)
 *
 * Inputs:
 * X -- nP by ncolX matrix of coded genotype of all children that are
 *      possible in the nuclear family. 
 * ncolX is the dimension of a coded genotype.
 * Prob -- conditional probability of child's genotype
 * nP is the number of genotypes that 
 *      are possible in the nuclear family; 
 * sumTi -- sum_{j}^{nChildren}T[j]
 * sumTi2 -- sum_{j}^{nChildren}T[j]^2
 *
 * Outpts:
 * CovSi -- ncolX by ncolX conditional covariance matrix of Si.
 *        Cov(Si)=(sum_{j}^{nChildren}T[j])*part1+
 *                (sum_{j}&{nChildren}T[j]^2*part2
 *        part1=sum_{g}^{G}sum_{g'}^{G}{(P[g,g']-P[g]*P[g'])X[g]X[g']^t}
 *        part2=sum_{g}^{G}P[g]*X[g]*X[g]^t-
 *          sum_{g}^{G}sum_{g'}^{G}{P[g,g']X[g]X[g']^t}
 */
void getCovSiFun(double *X, int *ncolX, double *Prob, int *nP, double *sumTi,
		double *sumTi2, double *jointProb, double *CovSi)
{
  int j, g, gt, G, nncolX, one, pos, tmpg, tmpgt;
  double sumProb, *part1, *part21, *part22, *part2, *Xg, *tXg;
  double *Xgt, *tXgt, zero, tmp;

  G=*nP;
  nncolX=*ncolX;
  zero=0.0;
  one=1;

  if(G<2)
  { matIni(CovSi, &zero, &nncolX, &nncolX); 
    return; 
  }
	
  part1=(double *)gS_alloc((size_t)(nncolX*nncolX), sizeof(double));
  part21=(double *)gS_alloc((size_t)(nncolX*nncolX), sizeof(double));
  part22=(double *)gS_alloc((size_t)(nncolX*nncolX), sizeof(double));
  part2=(double *)gS_alloc((size_t)(nncolX*nncolX), sizeof(double));
  Xg=(double *)gS_alloc((size_t)(nncolX), sizeof(double));
  tXg=(double *)gS_alloc((size_t)(nncolX), sizeof(double));
  Xgt=(double *)gS_alloc((size_t)(nncolX), sizeof(double));
  tXgt=(double *)gS_alloc((size_t)(nncolX), sizeof(double));

  sumProb=0.0;
  for(g=0; g<G; g++)
  { sumProb += Prob[g]; }
  if(fabs(sumProb-1.0)>1.0e-6)
  { error("Sum of all probabilities (%lf) not equal to 1!", sumProb);
    //free(part1); free(part21); free(part22); free(part2); free(Xg); // Automatically performed for 'S_alloc'
    //free(tXg); free(Xgt); free(tXgt); // Automatically performed for 'S_alloc'
  }

  //part21=sum_{g}^{G}P[g]*X[g]*X[g]^t
  matIni(part21, &zero, ncolX, ncolX);
  for(g=0; g<G; g++)
  { for(j=0; j<nncolX; j++)
    { Xg[j]=X[g*nncolX+j]; }

    // part21 = part21 + Prob[g]*Xg*Xg^t
    matTranspose(Xg, ncolX, &one, tXg);
    sumaAB(Xg,  ncolX, &one, tXg, ncolX, &Prob[g], part21);
  }

  // part22=sum_{g}^{G}sum_{g'}^{G}{P[g,g']X[g]X[g']^t}
  matIni(part22, &zero, ncolX, ncolX);
  //part1=sum_{g}^{G}sum_{g'}^{G}{(P[g,g']-P[g]*P[g'])X[g]X[g']^t}
  matIni(part1, &zero, ncolX, ncolX);
  for(g=0; g<G; g++)
  { for(j=0; j<nncolX; j++)
    { Xg[j]=X[g*nncolX+j]; }

    for(gt=0; gt<G; gt++)
    { 
      for(j=0; j<nncolX; j++)
      { Xgt[j]=X[gt*nncolX+j]; }

      tmpg=g+1;
      tmpgt=gt+1;

      if(g<gt)
      { pos=G*(tmpg-1)-(tmpg-1)*(tmpg-2)/2+(tmpgt-tmpg); 
      }
      else  // j>k
      { 
        pos=G*(tmpgt-1)-(tmpgt-1)*(tmpgt-2)/2+(tmpg-tmpgt); 
      }
      // part22 = part22 + jointProb[g, g']*Xg*Xg^t
      matTranspose(Xgt, ncolX, &one, tXgt);
      sumaAB(Xg, ncolX, &one, tXgt, ncolX, 
        	      &jointProb[pos], part22); 
      
      tmp=jointProb[pos]-Prob[g]*Prob[gt];
      sumaAB(Xg,  ncolX, &one, tXgt, ncolX, 
          	      &tmp, part1); 
    }
  }
  // part2=part1-part2
  matSub(part21, ncolX, ncolX, part22, part2);
  tmp=(*sumTi)*(*sumTi);
  matScalar(part1, &tmp, ncolX, ncolX);
  matScalar(part2, sumTi2, ncolX, ncolX);
  matAdd(part1, ncolX, ncolX, part2, CovSi); 

  //free(part1); free(part21); free(part22); free(part2); free(Xg); // Automatically performed for 'S_alloc'
  //free(tXg); free(Xgt); free(tXgt); // Automatically performed for 'S_alloc'
  return;
}

	
/*
 * given genotypes at a marker in a nuclear family, get the children genotype
 * configuration
 *
 * Inputs:
 * children -- nChildren by 2 matrix of genotypes of children
 * nChildren -- number of children
 * 
 * Outputs:
 * nType -- number of different genotypes of children genotype
 * config -- offspring genotype configuration, i.e., the set of
 *           different genotypes of children genotype
 */
int *childrenConfig(int *children, int *nChildren, int *nType, int *nAA,
		int *nAB)
{
  int i, j, k, *config, nnChildren, ncol, nnType, flag;
  int nnAA, nnAB, tmpcount, *alleles, nAllele;

  //count number of different alleles
  alleles=countAlleles(children, nChildren, &nAllele);
  if(nAllele>4) // number of different alleles more than 4 
  {
    printf("Error! number of different children alleles in a nuclear family> 4! Impossible!\n");
    printf("program terminated!\n");
    return(NULL);
  }
  //free(alleles); // Automatically performed for 'S_alloc'
  
  nnChildren=*nChildren;
  ncol=2;
  nnAA=0;
  nnAB=0;

  //there are at most nnChildren different observed offspring genotype
  config=(int *)gS_alloc((size_t)(nnChildren*ncol), sizeof(int));

  for(j=0; j<ncol; j++)
  { config[0*ncol+j]=children[0*ncol+j]; }
  if(children[0*ncol+0]==children[0*ncol+1])
  { nnAA++; } // number of homozygous genotypes AA or aa
  else { nnAB++; } // number of heterozygous genotypes Aa or aA

  nnType=1;
		  
  for(i=1; i<nnChildren; i++)
  { if(children[i*ncol+0]==children[i*ncol+1])
    { nnAA++; }
    else {nnAB++;}

    tmpcount=0;
    //if tmpsum is not equal to all the previous sum of two alleles of child,
    //then this child's genotype is a new type
    for(k=0; k<nnType; k++)
    { flag=checkEqualGeno((children+i*ncol), (config+k*ncol));
      if(flag!=0)  //equal
      { continue; }
      else // not equal
      { tmpcount++; } 
    }
    if(tmpcount==nnType) // new genotype
    { config[nnType*ncol+0]=children[i*ncol+0];
      config[nnType*ncol+1]=children[i*ncol+1];
      nnType ++;
    }
  }

  // delete unused memory if nnType<nChidren
  if(nnType<nnChildren)
  { config=(int *)gS_realloc(config, 
                            (size_t) (nnType*ncol),
                            (size_t) (nnChildren*ncol),
                            (size_t) sizeof(int));
  }
  if(nnType>4) //impossible
  { printf("Error! The number of possible children genotypes > 4! Impossible!\n");
    printf("Program terminated!\n");
    return(NULL);
  }

  
  *nType=nnType;
  *nAA=nnAA;
  *nAB=nnAB;
  return(config);
}

/*
 * get all possible offspring genotypes that are possible in the family
 *
 * Input:
 * parents -- nParents by 2 matrix of parental genotypes
 * nParents -- number of parents whose genotpes are availabe
 * children -- nChildren by 2 matrix of chidren genotypes
 * nChildren -- number of children in the nuclear family
 *
 * Output:
 * nPType --  number of possible offspring genotypes
 * pconfig -- possible offspring genotype configuration
 */
int *possibleGeno(int *parents, int *nParents, int *children, int *nChildren,
		int *nPType, int *compatibleFlag)
{
  int *config, i, nType, nAA, nAB, nnParents, nnChildren, *pconfig, ncol;
  int flag, *alleles, nAllele, oldsize;
  int tmpconfig[4*2], tmpn, p1, p2, q1, q2;

  nnParents=*nParents;
  nnChildren=*nChildren;
  ncol=2;
  pconfig=NULL;

  *compatibleFlag=1; 
  // obtain children's genotype configuration,i.e., set of different 
  // offspring genotypes observed.
  config=childrenConfig(children, nChildren, &nType, &nAA, &nAB);
  if(config == NULL)
  { *compatibleFlag=0;
    return(pconfig);
  }

  //check if parental genotypes are compatible with children's genotypes
  flag=1;
  if(nnParents==2)
  { for(i=0; i< nType; i++)
    { checkCompatibility1(parents, (config+i*2), &flag);
      if(flag==0) // not compatible
      { break; }	
    }
  } else if(nnParents==1)
  { for(i=0; i< nType; i++)
    { checkCompatibility2(parents, (config+i*2), &flag);
      if(flag==0) // not compatible
      { break; }	
    }
  }
  if(flag==0) // not compatible
  { *compatibleFlag=0;
    //free(config); // Automatically performed for 'S_alloc'
    return(pconfig);
  }	

  //count number of different alleles
  alleles=countAlleles(config, &nType, &nAllele);
  if(nAllele>2)
  { //printf("config>>\n"); myPrintMatI(config, nAllele, 2);
    printf("Current version allows two different alleles only!\n");
    printf("program terminated!\n");
    *compatibleFlag=0;
    //free(config); // Automatically performed for 'S_alloc'
    return(pconfig);
  }
  /////// end of checking compatibility

  if(nnParents==2)
  {
    p1=parents[0*2+0]; p2=parents[0*2+1];
    q1=parents[1*2+0]; q2=parents[1*2+1];
    tmpconfig[0*2+0]=p1; tmpconfig[0*2+1]=q1;
    tmpconfig[1*2+0]=p1; tmpconfig[1*2+1]=q2;
    tmpconfig[2*2+0]=p2; tmpconfig[2*2+1]=q1;
    tmpconfig[3*2+0]=p2; tmpconfig[3*2+1]=q2;

    tmpn=4;
    pconfig=childrenConfig(tmpconfig, &tmpn, nPType, &nAA, &nAB);
    if(pconfig == NULL)
    { *compatibleFlag = 0; // error occurs
      return(pconfig);
    }
  } else if (nnParents==1 && parents[0]==parents[1]) {
    pconfig=childrenConfig(children, nChildren, nPType, &nAA, &nAB);
    if(pconfig == NULL)
    { *compatibleFlag = 0; // error occurs
      return(pconfig);
    }
  } else {
    pconfig=childrenConfig(children, nChildren, nPType, &nAA, &nAB);
    if(pconfig == NULL)
    { *compatibleFlag = 0; // error occurs
      return(pconfig);
    }
    if(*nPType==2&&pconfig[0*2+0]==pconfig[0*2+1]
		    &&pconfig[1*2+0]==pconfig[1*2+1])
    { //e.g. {AA, BB} 
      oldsize=*nChildren;
      if(*nPType<*nChildren)
      { oldsize=*nPType; }
      pconfig=(int *)gS_realloc(pconfig, 
                               (size_t) (3*2),
                               (size_t) (oldsize*2),
                               (size_t) sizeof(int));  
      pconfig[2*2+0]=alleles[0];
      pconfig[2*2+1]=alleles[1];
      *nPType=3;
      //now pconfig= {AA, BB, AB}
    }
  }
  return(pconfig);

  //free(alleles); // Automatically performed for 'S_alloc'
  return(pconfig); 
}


/* for one nuclear family one marker */
/*
 * Inputs:
 *
 * parentsiM -- parents' genotypes at marker M in i-th family 
 * nParentsiM -- number of parents with no missing alleles
 * childreniM -- children's genotyes at marker M in i-th family
 * nChildreniM -- number of children with no missing alleles
 * allelesM -- alleles at marker M
 * nAlleleM -- number of alleles at marker M
 * genoMethod -- genotype coding method
 * nX -- number of elements of score vector
 * Tvec -- coded traits
*
 * Outputs:
 * SiM -- 1 by nX vector. 
 * ESiM -- 1 by nX vector. 
 * CovSim -- nX by nX vector. 
 */
void familyiM(int *parentsiM, int *nParentsiM, int *childreniM, 
		int *nChildreniM, int *allelesM, int *nAlleleM, 
		int *genoMethod, int *nX, double *Tvec, double *SiM, 
		double *ESiM, double *CovSiM, int *compatibleFlag)
{
  int i, j, k, nnParents, nnChildren, nPType, nchildType, nJoint, nAA, nAB; 
  int *pconfig, *configM, nnX, nnAllele, loop, tmpj, tmpk, pos, one;
  int *alleles2, nAllele2, flag, childi[2];
  double *prob, *jointProb, *Xmat, *XmatP, sumTiM, sumTi2M, zero;


  nnParents=*nParentsiM;
  nnChildren=*nChildreniM;
  nnAllele=*nAlleleM;
  nnX=*nX;
  zero=0.0;
  one=1;
  *compatibleFlag=1; //compatbile
  
  ////////// check compatibility
  alleles2=countAlleles(childreniM, nChildreniM, &nAllele2);
  if(nAllele2>2)
  { printf("Current version allows two different alleles only!\n");
    printf("program terminated!\n");
    *compatibleFlag = 0;
    return;
  }
  //free(alleles2); // Automatically performed for 'S_alloc'
  flag=1;
  if(nnParents==2)
  { for(i=0; i< nnChildren; i++)
    { childi[0*2+0]=childreniM[i*2+0];
      childi[0*2+1]=childreniM[i*2+1];
      checkCompatibility1(parentsiM, childi, &flag);
      if(flag==0) // not compatible
      { *compatibleFlag=0;
        break; 
      }	
    }
  } else if(nnParents==1)
  { for(i=0; i< nnChildren; i++)
    { childi[0*2+0]=childreniM[i*2+0];
      childi[0*2+1]=childreniM[i*2+1];
      checkCompatibility2(parentsiM, childi, &flag);
      if(flag==0) // not compatible
      { *compatibleFlag=0;
        break; 
      }	
    }
  }
  if(flag==0) // not compatible
  { matIni(SiM, &zero, &one, nX);
    matIni(ESiM, &zero, &one, nX);
    matIni(CovSiM, &zero, nX, nX);
    return;
  }	
  ///////////// end of checking compatibility
  
  //records coded genotype for children
  Xmat=(double *)gS_alloc(nnChildren*nnX, sizeof(double));

  // get coded observed genotype nChildren by nX matrix at marker M
  for(j=0; j<nnChildren; j++)
  { Xfun((childreniM+j*2), allelesM, nAlleleM, genoMethod, nX,
      	    (Xmat+j*nnX)); }
  
  //obtain scores Si at marker M
  getSiFun(Xmat, nX, Tvec, nChildreniM, SiM, &sumTiM, &sumTi2M);

  // ********* calculate ESi and CovSi at marker M *************
  
  // first, obtain all possible offspring genotypes

  pconfig=possibleGeno(parentsiM, nParentsiM, childreniM, nChildreniM,
    	&nPType, compatibleFlag); // there are nPType possible offspring genotype
  if(*compatibleFlag==0)
  {
    matIni(SiM, &zero, &one, nX);
    matIni(ESiM, &zero, &one, nX);
    matIni(CovSiM, &zero, nX, nX);
    return;
  }


  /*
   * for 1<=j<=k<=n, joint prob between (row) j and k is
   * jointProb[n*(j-1)-(j-1)*(j-2)/2+(k-j)]
   */
  nJoint=nPType*(nPType+1)/2;

  // nPType=1 means informative family count is 0
  if(nPType>1)
  { jointProb=(double *)gS_alloc((size_t) (nJoint), sizeof(double)); 
    // obtain the conditional probabilities for children at marker M
    prob=(double *)gS_alloc((size_t) (nPType), sizeof(double));

    if(nnParents==2)  // complete parental genotype
    {
      //obtain conditional prob for j-th possible child
      for(j=0; j<nPType; j++)
      { condProbCPj(parentsiM, (pconfig+j*2), (prob+j), compatibleFlag); }
      loop=0;
      //obtain joint conditional prob for j-th and k-th children
      for(j=0; j<nPType; j++)
      { for(k=j; k<nPType; k++)
        { jointProb[loop]=prob[j]*prob[k]; 
          loop ++;
        }
      }
    } //end if
    else { // at least one parent's genotype is missing
      // obtain the genotype configuration of children in the family
      configM=childrenConfig(childreniM, nChildreniM, &nchildType, 
        &nAA, &nAB);
      if(pconfig == NULL)
      { *compatibleFlag = 0; // error occurs
        return;
      }

      for(j=0; j<nPType; j++)
      { for(k=j; k<nPType; k++)
        { tmpj=j+1;
          tmpk=k+1;
          pos=nPType*(tmpj-1)-(tmpj-1)*(tmpj-2)/2+(tmpk-tmpj);
          // calculate joint conditional probability
          if(nnParents==1 && parentsiM[0]==parentsiM[1])
          // one parent missing, and the other parent is homozygous
          { condProbM1PHomojk(nChildreniM, &nchildType, 
              (pconfig+j*2), (pconfig+k*2), (prob+j), (prob+k), 
              (jointProb+pos));
            if(prob[j] < 0)
            { *compatibleFlag = 0;
              return;
            }
          } else {
           // nnParents=1, parentsiM[0]!=parentsiM[1], or nnParents=0
           // i.e. one parent missing, the other parent heterozygous
           // or both parents are missing
            condProbMPHeterojk(nChildreniM, &nchildType, configM, 
      		&nAA, &nAB, (pconfig+j*2), (pconfig+k*2), (prob+j), 
      		(prob+k), (jointProb+pos));
            if(prob[j] < 0) // something wrong
            { *compatibleFlag=0;
              return;
            }
          }
        }
      }
    } //end if
 
    // coded genotype for all possible children genotypes
    XmatP=(double *)gS_alloc((size_t) (nPType*nnX), sizeof(double));
    // get all possible coded genotype (nType by nX matrix)
    for(j=0; j<nPType; j++)
    { Xfun((pconfig+j*2), allelesM, nAlleleM, genoMethod, nX,
      	    (XmatP+j*nnX)); }
 
    //get ESi at marker M
    getESiFun(XmatP, nX, prob, &nPType, &sumTiM, ESiM);
 
    //get CovSi at marker M
    getCovSiFun(XmatP, nX, prob, &nPType, &sumTiM, &sumTi2M,
        	      jointProb, CovSiM);
    //free(jointProb); free(XmatP); free(prob);  // Automatically performed for 'S_alloc'
    //if(nnParents!=2) { free(configM); } // Automatically performed for 'S_alloc'
  } else { // nPType = 1
    matIni(SiM, &zero, &one, nX);
    matIni(ESiM, &zero, &one, nX);
    matIni(CovSiM, &zero, nX, nX);
  }

  //free(Xmat); free(pconfig);  // Automatically performed for 'S_alloc'

  return;
}

/*
 * Given a marker and n nuclear families, do FBAT test.
 *
 * pedM -- nrow by 8 matrix of pedigrees of n nuclear families, where 
 *        nrow=sum_{i=1}^{nFamily} familySize[i]. The first 6 columns are pid, id, 
 *        fid, mid, sex, aff. And the remaining 2 columns are a pair 
 *        of marker alleles.
 * marker -- indicate which marker we are dealing with (starting with 0)
 * nFamily -- number of nuclear families
 * familySize -- nFamily by 1 vector of sizes of nuclear families. Size of a nuclear
 *      family is the number of individuals in this family.
 * genoMethod -- genotype coding method
 * nX -- number of elements of score vector
 * traitMethod -- trait coding method
 * traitOffset -- offset for the trait
 * alleles -- alleles of the marker
 * nAllele -- number of alleles of the marker
 *
 * Outputs:
 * SM -- a 1 by nX matrix. SM=sum_{i=1}^{nFamily} SMi, where 
 *       SMi=sum_{j}^{nChildreni} X[j]*T[j], 
 *   where X[j] and T[j] are the coded genotype and trait, respectively, 
 *   for the j-th children. 'nX'='nAllelesM' or 'nAllelesM*(nAllelesM+1)/2-1'
 * ESM -- a 1 by nX matrix of conditional expectations of SM
 *       'nX'='nAllelesM' or 'nAllelesM*(nAllelesM+1)/2-1' 
 * CovSM -- a nX by nX array of conditional covariance of SM
 *       'nX'='nAllelesM' or 'nAllelesM*(nAllelesM+1)/2-1' 
 */
void fbatM(int *pedM, int *marker, int *nFamily, int *familySize, int *genoMethod, 
		int *nX, int *traitMethod, double *traitOffset, 
		int *alleles, int *nAllele, double *SM, 
		double *ESM, double *CovSM, 
		int *numInfoM, int *quiet)
{
  int nnX, i, j, k, rr, nnFamily, nncol, one;
  int missingFamily, nnAllele, start, tmpflag;
  int *pedMi, numParents, numChildren; 
  int *childVec, tmpfamilysize;
  double zero;

  nnFamily=*nFamily; // number of nuclear families
  nncol=8;
  zero=0.0;
  one=1;
  nnAllele=*nAllele;
  nnX=*nX;

  // initialize SM, ESM, CovSM with zero
  matIni(SM, &zero, &one, nX);
  matIni(ESM, &zero, &one, nX);
  matIni(CovSM, &zero, nX, nX);

  start=0;
  missingFamily=0;
  *numInfoM=0;
  for(i=0; i<nnFamily; i++)
  {
    /* get ped for family i at marker M */
    numParents=0;
    numChildren=0;
    pedMi=(int *)gS_alloc((size_t) familySize[i]*8, sizeof(int));
    childVec=(int *)gS_alloc((size_t) familySize[i], sizeof(int));
    tmpfamilysize=0;
    for(j=0; j<familySize[i]; j++)
    { rr=start+j;
      tmpflag=0;
      for(k=0; k<8; k++)
      { pedMi[tmpfamilysize*8+k]=pedM[rr*8+k]; 
	if((k==3)&& (pedMi[tmpfamilysize*8+k]!=0) && (pedMi[tmpfamilysize*8+k-1]!=0)) 
	{ childVec[numChildren]=tmpfamilysize;
	  numChildren++; 
	}
	else if((k==3)&& (pedMi[tmpfamilysize*8+k]==0) && (pedM[tmpfamilysize*8+k-1]==0)) 
	{ numParents++; }
	else if((k==3)&& (((pedMi[tmpfamilysize*8+k]==0) && (pedMi[tmpfamilysize*8+k-1]!=0)) ||
			((pedMi[tmpfamilysize*8+k]!=0)&&(pedMi[tmpfamilysize*8+k-1]==0))))
	{  printf("could not determine if the individual with id %d in Family with pid %d at %d-th marker is a parent or a child! This individual will be exluded from the analysis!\n", pedMi[tmpfamilysize*8+1], pedMi[tmpfamilysize*8+0], (*marker+1));
	   // set this individual as parents and set alleles as missing
	   tmpflag=1;
	   break; // terminate the loop for(k=0; k<8; k++)
        }
      }
      if(tmpflag==0){
        tmpfamilysize++;
      }
    }
    if(tmpfamilysize==0) //this family is not informative. continue next family
    { //free(pedMi); free(childVec); // Automatically performed for 'S_alloc'
      continue; 
    }
    if(tmpfamilysize<familySize[i])
    { pedMi=(int *)gS_realloc(pedMi, 
                             (size_t) (tmpfamilysize*8),
                             (size_t) familySize[i]*8,
                             (size_t) sizeof(int));
      familySize[i]=tmpfamilysize;
    }
    if(numChildren==0)
    { missingFamily++; } 
    else 
    { // decompose to several nuclear families
      decomposePed(marker, pedMi, &familySize[i], childVec, &numChildren, alleles, nAllele, genoMethod, traitMethod, traitOffset, nX, SM, ESM, CovSM, &missingFamily, numInfoM, quiet);
    }
    start += familySize[i];
    //free(childVec); free(pedMi);  // Automatically performed for 'S_alloc'
  }
  /**
  if(missingFamily==nnFamily) { 
    //all children genotypes in all family are missing
    if(*iprint!=0){
      warning("Warning! All children's genotypes at marker %d in all Family are missing!\n", (*marker+1));
    }
  }
  **/

  return;
}


/*************************************************************************/

void checkCompatible(int *ped, int *nrow, int *ncol, int *nFamily, 
                     int *familySize, int *alleles, int *nAllele, 
                     int *sampleInfoLen, int *compatibleFlag, 
                     int *errorFlag, int *whichMarker, 
                     int *nMerrMarker, int *nMerrFamily, 
                     int *nErrFamilySample, int *quiet)
{
  int i, j, M, nnFamily, nnrow, nncol, nMarker;
  int *pedM, start, end, nnAllele; 
  int *tmpAlleles, start2, end2, nSampleInfoLen, nc2;
  double zero; 
  

  start2=0;
  *errorFlag=0;
  nSampleInfoLen=*sampleInfoLen;
  nc2=nSampleInfoLen+2;
  nnFamily=*nFamily; // number of nuclear families
  nnrow=*nrow; // number of individuals in the pedigree file
  nncol=*ncol; // number of columns of the pedigree file
  nMarker=(nncol-nSampleInfoLen)/2; // number of markers
  zero=0.0;

  //input first 6 columns of pedigrees
  //For each marker, the pedigree contains 8 columns
  //First 6 columns are 'pid', 'id', 'fid', 'mid', 'sex', 'aff'.
  //The rest two columns are two alleles of the marker
  pedM=(int *)gS_alloc((size_t)nnrow*nc2, sizeof(int));
  for(i=0; i<nnrow; i++)
  { for(j=0; j<nSampleInfoLen; j++)
    { 
      pedM[i*nc2+j]=ped[i*nncol+j]; 
    }
    //fid=id or mid=id
    if(pedM[i*nc2+2]==pedM[i*nc2+1])
    {
      if(*quiet==0)
      { warning("Warning! father id = subject id = %d!", pedM[i*nc2+1]); }
      *errorFlag=3;
      nErrFamilySample[i] ++;
      //free(pedM); // Automatically performed for 'S_alloc'
      return;
    }
    if(pedM[i*nc2+3]==pedM[i*nc2+1])
    {
      if(*quiet==0)
      { warning("Warning! mother id = subject id = %d!", pedM[i*nc2+1]); }
      *errorFlag=3;
      nErrFamilySample[i] ++;
      //free(pedM); // Automatically performed for 'S_alloc'
      return;
    }

  }

  //for each marker do FBAT analysis
  for(M=0; M<nMarker; M++)
  {
    //input the genotype info
    start=nSampleInfoLen+M*2;
    end=nSampleInfoLen+M*2+1;
    for(i=0; i<nnrow; i++)
    { pedM[i*nc2+nSampleInfoLen]=ped[i*nncol+start]; 
      pedM[i*nc2+nSampleInfoLen+1]=ped[i*nncol+end]; 
    }

    //create temporary score, expected score and covariance matrix
    //Note that for different markers, the number of alleles may be different
    //'allele' is a vector consecutively recording the alleles of all markers 

    nnAllele=nAllele[M]; // number of alleles for this marker
    
    //get the alleles for this marker
    tmpAlleles=(int *)gS_alloc((size_t) nnAllele, sizeof(int));
    if(M>0)
    { end2=start2+nAllele[M]; 
    } else {
      start2=0;
      end2=nAllele[M];
    }
    j=0;
    for(i=start2; i<end2; i++)
    { tmpAlleles[j]=alleles[i]; 
      j++; 
    }
    
    start2=end2;

    //do FBAT analysis
    checkCompatibleM(pedM, &M, nFamily, familySize, tmpAlleles, 
                     &nnAllele, sampleInfoLen, compatibleFlag,
                     nMerrMarker, nMerrFamily, quiet); 
  }
  //free(pedM); // Automatically performed for 'S_alloc'

  return;
}


void checkCompatibleM(int *pedM, int *marker, int *nFamily, int *familySize, 
		int *alleles, int *nAllele, int *sampleInfoLen, 
                int *compatibleFlag, 
                int *nMerrMarker, int *nMerrFamily,
                int *quiet)
{
  int i, j, k, rr, nnFamily, nncol, one;
  int missingFamily, nnAllele, start, tmpflag;
  int *pedMi, numParents, numChildren; 
  int *childVec, tmpfamilysize, nSampleInfoLen;
  int flagMerrMarkerMi;
  double zero;

  nnFamily=*nFamily; // number of nuclear families
  nSampleInfoLen=*sampleInfoLen;
  nncol=nSampleInfoLen+2;
  zero=0.0;
  one=1;
  nnAllele=*nAllele;

  start=0;
  missingFamily=0;
  for(i=0; i<nnFamily; i++)
  {
    flagMerrMarkerMi=0; // Mendelian's error for marker 'M' & family 'i'
    /* get ped for family i at marker M */
    numParents=0;
    numChildren=0;
    pedMi=(int *)gS_alloc((size_t) familySize[i]*nncol, sizeof(int));
    childVec=(int *)gS_alloc((size_t) familySize[i], sizeof(int));
    tmpfamilysize=0;
    for(j=0; j<familySize[i]; j++)
    { rr=start+j;
      tmpflag=0;
      for(k=0; k<nncol; k++)
      { pedMi[tmpfamilysize*nncol+k]=pedM[rr*nncol+k]; 
	if((k==3)&& (pedMi[tmpfamilysize*nncol+k]!=0) && (pedMi[tmpfamilysize*nncol+k-1]!=0)) 
	{ childVec[numChildren]=tmpfamilysize;
	  numChildren++; 
	}
	else if((k==3)&& (pedMi[tmpfamilysize*nncol+k]==0) && (pedM[tmpfamilysize*nncol+k-1]==0)) 
	{ numParents++; }
	else if((k==3)&& (((pedMi[tmpfamilysize*nncol+k]==0) && (pedMi[tmpfamilysize*nncol+k-1]!=0)) ||
			((pedMi[tmpfamilysize*nncol+k]!=0)&&(pedMi[tmpfamilysize*nncol+k-1]==0))))
	{  if(*quiet==0) 
           {
	     warning("Warning! Could not determine if the individual with id %d in Family with pid %d at %d-th marker is a parent or a child! This individual will be exluded from the analysis!", pedMi[tmpfamilysize*nncol+1], pedMi[tmpfamilysize*nncol+0], (*marker+1));
	   }
           flagMerrMarkerMi = 1;
           nMerrMarker[*marker] += flagMerrMarkerMi;
           nMerrFamily[i] += flagMerrMarkerMi;
           *compatibleFlag = 0;
	   // set this individual as parents and set alleles as missing
	   tmpflag=1;
	   break; // terminate the loop for(k=0; k<nncol; k++)
        }
      }
      if(tmpflag==0){
        tmpfamilysize++;
      }
    }
    if(tmpfamilysize==0) //this family is not informative. continue next family
    { //free(pedMi); free(childVec); // Automatically performed for 'S_alloc'
      continue; 
    }
    if(tmpfamilysize<familySize[i])
    { pedMi=(int *)gS_realloc(pedMi, 
                             (size_t) tmpfamilysize*nncol, 
                             (size_t) familySize[i]*nncol, 
                             (size_t) sizeof(int));
     familySize[i]=tmpfamilysize;
    }
    if(numChildren==0)
    { missingFamily++; } 
    else 
    { // decompose to several nuclear families
      checkdecomposePed(marker, pedMi, &familySize[i], childVec, 
                        &numChildren, alleles, nAllele, sampleInfoLen, 
                        &missingFamily, compatibleFlag, 
                        &flagMerrMarkerMi, quiet);
    }

    start += familySize[i];
    //free(childVec); free(pedMi);  // Automatically performed for 'S_alloc'

    nMerrMarker[*marker] += flagMerrMarkerMi;
    nMerrFamily[i] += flagMerrMarkerMi;
  }
  return;
}

void checkdecomposePed(int *marker, int *pedMi, int *familySize, 
                       int *childVec, int *numChildren, int *alleles, 
                       int *nAllele, int *sampleInfoLen, int *missingFamily, 
                       int *compatibleFlag, 
                       int *flagMerrMarkerMi, int *quiet)
{
  int i, j, k, nNumChildren, nFamilySize, nLeft, tmpNchild, tmpNparent; 
  int tmpNfamily, pos, fid, mid;
  int *tmpPedMi, nSampleInfoLen, nc2;

  nNumChildren=*numChildren;
  nFamilySize=*familySize;
  nLeft=nNumChildren;
  nSampleInfoLen=*sampleInfoLen;
  nc2=nSampleInfoLen+2;
  i=0;

  do {
    for(j=0; j<nNumChildren; j++)
    { if(childVec[j]!=-1)
      { i=j;
        break;
      }
    }
    tmpPedMi=(int *)gS_alloc((size_t) nFamilySize*nc2, sizeof(int));
    pos=childVec[i];
    for(k=0; k<nc2; k++)
    { tmpPedMi[0*nc2+k]=pedMi[pos*nc2+k]; }
    childVec[i]=-1;
    tmpNchild=1;
    //find children having the same parents as the current children
    for(j=0; j<nNumChildren; j++)
    { if(childVec[j]!= -1)
      { //two children have the same parents
	pos=childVec[j];
        if(pedMi[pos*nc2+2]==tmpPedMi[0*nc2+2] &&
           pedMi[pos*nc2+3]==tmpPedMi[0*nc2+3]) 
        { 
	  for(k=0; k<nc2; k++)
	  { tmpPedMi[tmpNchild*nc2+k]=pedMi[pos*nc2+k]; }
          tmpNchild++; 
	  childVec[j]=-1;
        }
      }
    }
    // get parents info for this nuclear family
    fid=tmpPedMi[0*nc2+2];
    mid=tmpPedMi[0*nc2+3];
    tmpNparent=0;
    for(j=0; j<nFamilySize; j++)
    { if((pedMi[j*nc2+2]==0&&pedMi[j*nc2+3]==0)&&(fid==pedMi[j*nc2+1] || mid==pedMi[j*nc2+1]))
      { pos=tmpNchild+tmpNparent;
        for(k=0; k<nc2; k++)
        { tmpPedMi[pos*nc2+k]=pedMi[j*nc2+k]; }
	tmpNparent++;
      } else if((pedMi[j*nc2+2]!=0||pedMi[j*nc2+3]!=0)&&(fid==pedMi[j*nc2+1] || mid==pedMi[j*nc2+1]))
      {
	if(*quiet==0)
	{ warning("Warning! The family with id=%d is not a nuclear family!", pedMi[j*nc2+0]);
        }
        *flagMerrMarkerMi = 1;
        *compatibleFlag = 0;
      }
    }
    tmpNfamily=tmpNchild+tmpNparent;
    tmpPedMi=(int *)gS_realloc(tmpPedMi, 
                              (size_t) tmpNfamily*nc2,
                              (size_t) nFamilySize*nc2,
                              (size_t) sizeof(int));
    // for one nuclear family
    checkoneNuclearFamily(marker, tmpPedMi, &tmpNfamily, alleles, 
                          nAllele, sampleInfoLen, missingFamily, 
                          compatibleFlag, flagMerrMarkerMi,
                          quiet); 
    nLeft -= tmpNchild;
  } while(nLeft>0); 
  return;
}

void checkoneNuclearFamily(int *marker, int *pedMi, int *familySize, 
                           int *alleles, int *nAllele, int *sampleInfoLen, 
                           int *missingFamily, int *compatibleFlag, 
                           int *flagMerrMarkerMi, int *quiet)
{
  int j, missingChild, nncol, sex1, sex2, nParentsi, nChildreni;
  int *parentsi, *childreni, nfamilySize, nSampleInfoLen;
  int tmpCompatibleFlag;
  double zero;

  zero=0.0;
  nfamilySize=*familySize;
  nSampleInfoLen=*sampleInfoLen;
  nncol=nSampleInfoLen+2;
  sex1=0;
  sex2=0;

  tmpCompatibleFlag=1;
  
  parentsi=(int *)gS_alloc((size_t) (2*2),sizeof(int)); 
  childreni=(int *)gS_alloc((size_t) (nfamilySize*2),sizeof(int));

  nParentsi=0;  // number of parents in this nuclear family
  nChildreni=0; // number of children in this nuclear family

  // get pedigree for the i-th nuclear family
  for(j=0; j<nfamilySize; j++)
  { missingChild=0;
    if(pedMi[j*nncol+2]==0&&pedMi[j*nncol+3]==0) //parent
    { if(nParentsi==0) { sex1=pedMi[j*nncol+4]; } 
      else if(nParentsi==1) { sex2=pedMi[j*nncol+4]; } 
      if(pedMi[j*nncol+nSampleInfoLen]!=0&&pedMi[j*nncol+nSampleInfoLen+1]!=0) //not missing
      { 
        parentsi[nParentsi*2+0]=pedMi[j*nncol+nSampleInfoLen];
        parentsi[nParentsi*2+1]=pedMi[j*nncol+nSampleInfoLen+1];
        nParentsi++; 
      }
      //else { missing=1; }
    }
    else if(pedMi[j*nncol+2]!=0&&pedMi[j*nncol+3]!=0) //child
    { //check if chlid genotype is missing
      if(pedMi[j*nncol+nSampleInfoLen]!=0&&pedMi[j*nncol+nSampleInfoLen+1]!=0) //not missing
      { 
        childreni[nChildreni*2+0]=pedMi[j*nncol+nSampleInfoLen];
        childreni[nChildreni*2+1]=pedMi[j*nncol+nSampleInfoLen+1];
        nChildreni++; 
      }
      else { 
        missingChild=1;
      }
    }
    else {//error!
      if(*quiet==0) 
      { warning("Warning! cannot deterimine if the individual with id %d in Family with pid %d at %d-th marker is parent or child! This individual will be exluded from analysis!", pedMi[0*nncol+1], pedMi[0*nncol+0], (*marker+1));
      }
      *flagMerrMarkerMi=1;
      *compatibleFlag=0;
    }
  }

  if(nParentsi==2)
  { //two parents have the same sex
    if(sex1==sex2)
    { if(*quiet==0) {
        warning("Warning! Inconsistent parental sex in family with pid %d !",
      		pedMi[0*nncol+0]);
        warning("This parents will be excluded from the analysis!\n");
      }
      *flagMerrMarkerMi=1;
      *compatibleFlag=0;
      nParentsi=0;
   }
  }

  //calculate Si, ESi, CovSi for each nuclear family
  if(nChildreni>0) 
  {
    if(nChildreni<nfamilySize)
    { childreni=(int *)gS_realloc(childreni, 
                                 (size_t) nChildreni*2,
                                 (size_t) (nfamilySize*2),
                                 (size_t) sizeof(int));
    }
    if(nParentsi==1)
    { parentsi=(int *)gS_realloc(parentsi, 
                                (size_t) (nChildreni*2),
                                (size_t) (2*2),
                                (size_t) sizeof(int));
    }
    checkfamilyiM(parentsi, &nParentsi, childreni, &nChildreni, alleles, 
                  nAllele, &tmpCompatibleFlag);
    if(tmpCompatibleFlag==0)
    { *flagMerrMarkerMi = 1; 
      *compatibleFlag = 0;

      if(*quiet==0)
      { warning("Warning! parental genotypes are not compatible with childrens' genotype in family with pid %d at %d-th marker!", pedMi[0*nncol+0], (*marker+1));
      }
  
    } 
  } else { // all children genotypes are missing
    if(*quiet==0)
    { warning("Warning! all childrens' genotypes are missing in family with pid %d at %d-th marker!", pedMi[0*nncol+0], (*marker+1)); }
    (*missingFamily)++;
    *flagMerrMarkerMi = 1;
    *compatibleFlag = 0;
  }
  //free(parentsi); free(childreni);  // Automatically performed for 'S_alloc'
  return;
}

void checkfamilyiM(int *parentsiM, int *nParentsiM, int *childreniM, 
		int *nChildreniM, int *allelesM, int *nAlleleM, 
		int *compatibleFlag)
{
  int i, nnParents, nnChildren; 
  int nnAllele;
  int *alleles2, nAllele2, flag, childi[2];
  double zero;

  nnParents=*nParentsiM;
  nnChildren=*nChildreniM;
  nnAllele=*nAlleleM;
  zero=0.0;
  //*compatibleFlag=1; //compatbile
  
  ////////// check compatibility
  alleles2=countAlleles(childreniM, nChildreniM, &nAllele2);
  //free(alleles2); // Automatically performed for 'S_alloc'
  flag=1;
  if(nnParents==2)
  { for(i=0; i< nnChildren; i++)
    { childi[0*2+0]=childreniM[i*2+0];
      childi[0*2+1]=childreniM[i*2+1];
      checkCompatibility1(parentsiM, childi, &flag);
      if(flag==0) // not compatible
      { *compatibleFlag=0;
        break; 
      }	
    }
  } else if(nnParents==1)
  { for(i=0; i< nnChildren; i++)
    { childi[0*2+0]=childreniM[i*2+0];
      childi[0*2+1]=childreniM[i*2+1];
      checkCompatibility2(parentsiM, childi, &flag);
      if(flag==0) // not compatible
      { *compatibleFlag=0;
        break; 
      }	
    }
  }
  ///////////// end of checking compatibility

  return;
}

int checkInformative(double *Si, double *ESi, double *CovSi, int *nX)
{
  double diffS, sumCov;
  int i, j, nnX;

  nnX=*nX;

  diffS=0.0;
  sumCov=0.0;
  for(i=0; i<nnX; i++)
  { diffS += fabs(Si[i]-ESi[i]);
    if(fabs(CovSi[i*nnX+i])<1.0e-6) {
      Si[i]=0.0;
      ESi[i]=0.0;
    }
    for(j=0; j<nnX; j++)
    { sumCov += fabs(CovSi[i*nnX+j]); }
  }

  if(fabs(diffS)<1.0e-6 && fabs(sumCov)<1.0e-6) // noninformative
  { return(0); }
  return(1);
}

int checkParentsID(int *pedMi, int familySizei, int j)
{
  int i, flag1, flag2;

  flag1=0; flag2=0;
  for(i=0; i<familySizei; i++)
  {
    if(pedMi[j*8+2]!=pedMi[i*8+1]) // father id not in id
    { flag1++; }
    if(pedMi[j*8+3]!=pedMi[i*8+1]) // mother id not in id
    { flag2++; }
  }
  if(flag1==familySizei||flag2==familySizei)
  { return(1); }
  return(0);
}

void oneNuclearFamily(int *marker, int *pedMi, int *familySize, int *alleles, int *nAllele, int *genoMethod, int *traitMethod, double *traitOffset, int *nX, double *SM, double *ESM, double *CovSM, int *missingFamily, int *numInfoM, int *quiet)
{
  int j, missingChild, nncol, sex1, sex2, nParentsi, nChildreni;
  int *parentsi, *childreni, nfamilySize, info, one, nnX, compatibleFlag;
  double *Tvec, traiti, zero;
  double *tmpSi, *tmpESi, *tmpCovSi, *tmpout1, *tmpout2;

  nncol=8;
  one=1;
  zero=0.0;
  nfamilySize=*familySize;
  nnX=*nX;
  sex1=0;
  sex2=0;

  //record temporary scores, expected scores, and covariance matrices
  tmpSi=(double *)gS_alloc((size_t)nnX, sizeof(double));
  tmpESi=(double *)gS_alloc((size_t)nnX, sizeof(double));
  tmpCovSi=(double *)gS_alloc((size_t)nnX*nnX, sizeof(double));
  tmpout1=(double *)gS_alloc((size_t)nnX, sizeof(double));
  tmpout2=(double *)gS_alloc((size_t)nnX*nnX, sizeof(double));
  
  parentsi=(int *)gS_alloc((size_t) (2*2),sizeof(int)); 
  childreni=(int *)gS_alloc((size_t) (nfamilySize*2),sizeof(int));
  Tvec=(double *)gS_alloc((size_t) (nfamilySize),sizeof(double));  

  nParentsi=0;  // number of parents in this nuclear family
  nChildreni=0; // number of children in this nuclear family

  // get pedigree for the i-th nuclear family
  for(j=0; j<nfamilySize; j++)
  { missingChild=0;
    if(pedMi[j*nncol+2]==0&&pedMi[j*nncol+3]==0) //parent
    { if(nParentsi==0) { sex1=pedMi[j*nncol+4]; } 
      else if(nParentsi==1) { sex2=pedMi[j*nncol+4]; } 
      if(pedMi[j*nncol+6]!=0&&pedMi[j*nncol+7]!=0) //not missing
      { 
        parentsi[nParentsi*2+0]=pedMi[j*nncol+6];
        parentsi[nParentsi*2+1]=pedMi[j*nncol+7];
        nParentsi++; 
      }
      //else { missing=1; }
    }
    else if(pedMi[j*nncol+2]!=0&&pedMi[j*nncol+3]!=0) //child
    { //check if chlid genotype is missing
      if(pedMi[j*nncol+6]!=0&&pedMi[j*nncol+7]!=0) //not missing
      { 
        childreni[nChildreni*2+0]=pedMi[j*nncol+6];
        childreni[nChildreni*2+1]=pedMi[j*nncol+7];
        traiti=pedMi[j*nncol+5]; // trait 
        // coded trait
        Tfun(&traiti, traitOffset, traitMethod, &Tvec[nChildreni]);
        nChildreni++; 
      }
      else { 
        missingChild=1;
      }
    }
    else {//error!
      if(*quiet==0) {
        warning("Warning! cannot deterimine if the individual with id %d in Family with pid %d at %d-th marker is parent or child! This individual will be exluded from analysis!", pedMi[0*nncol+1], pedMi[0*nncol+0], (*marker+1));
      }
    }
  }

  if(nParentsi==2)
  { //two parents have the same sex
    if(sex1==sex2)
    { if(*quiet==0) {
        warning("Warning! Inconsistent parental sex in family with pid %d !",
      		pedMi[0*nncol+0]);
        warning("This parents will be excluded from the analysis!");
      }
      nParentsi=0;
   }
  }

  //calculate Si, ESi, CovSi for each nuclear family
  if(nChildreni>0) 
  {
    if(nChildreni<nfamilySize)
    { childreni=(int *)gS_realloc(childreni, 
                                 (size_t) (nChildreni*2),
                                 (size_t) (nfamilySize*2),
                                 (size_t) sizeof(int));
      Tvec=(double *)gS_realloc(Tvec, 
                               (size_t) nChildreni,
                               (size_t) nfamilySize,
                               (size_t) sizeof(double));
    }
    if(nParentsi==1)
    { parentsi=(int *)gS_realloc(parentsi, 
                                (size_t) (nChildreni*2),
                                (size_t) (2*2),
                                (size_t) sizeof(int));
    }
    familyiM(parentsi, &nParentsi, childreni, &nChildreni, alleles, nAllele,
      	      genoMethod, nX, Tvec, tmpSi, tmpESi, tmpCovSi, &compatibleFlag);
    if((compatibleFlag==0) && (*quiet==0))
    { warning("Warning! parental genotypes are not compatible with childrens' genotype in family with pid %d at %d-th marker!", pedMi[0*nncol+0], (*marker+1));
    }
	    

    //check if the family is informative
    info=checkInformative(tmpSi, tmpESi, tmpCovSi, nX);
    if(info==1) { // infomative
      *numInfoM = *numInfoM+1;
      // tmpout1=SM+tmpSi
      matAdd(SM, &one, nX, tmpSi, tmpout1);
      // SM=tmpout1
      matAssign(tmpout1, &one, nX, SM);
      // tmpout1=ESM+tmpESi
      matAdd(ESM, &one, nX, tmpESi, tmpout1);
      // ESM=tmpout1
      matAssign(tmpout1, &one, nX, ESM);
      // tmpout2=CovSM+tmpCovSi
      matAdd(CovSM, nX, nX, tmpCovSi, tmpout2);
      // CovSM=tmput2
      matAssign(tmpout2, nX, nX, CovSM);
    }
  } else { // all children genotypes are missing
    (*missingFamily)++;
  }
  //free(parentsi); free(childreni); free(Tvec); // Automatically performed for 'S_alloc'
  //free(tmpSi); free(tmpESi);
  //free(tmpCovSi); free(tmpout1); free(tmpout2);
  return;
}

void decomposePed(int *marker, int *pedMi, int *familySize, int *childVec, int *numChildren, int *alleles, int *nAllele, int *genoMethod, int *traitMethod, double *traitOffset, int *nX, double *SM, double *ESM, double *CovSM, int *missingFamily, int *numInfoM, int *quiet)
{
  int i, j, k, nNumChildren, nFamilySize, nLeft, tmpNchild, tmpNparent; 
  int tmpNfamily, pos, fid, mid;
  int *tmpPedMi;

  nNumChildren=*numChildren;
  nFamilySize=*familySize;
  nLeft=nNumChildren;

  i = 0;
  do {
    for(j=0; j<nNumChildren; j++)
    { if(childVec[j]!=-1)
      { i=j;
        break;
      }
    }
    tmpPedMi=(int *)gS_alloc((size_t) nFamilySize*8, sizeof(int));
    pos=childVec[i];
    for(k=0; k<8; k++)
    { tmpPedMi[0*8+k]=pedMi[pos*8+k]; }
    childVec[i]=-1;
    tmpNchild=1;
    //find children having the same parents as the current children
    for(j=0; j<nNumChildren; j++)
    { if(childVec[j]!= -1)
      { //two children have the same parents
	pos=childVec[j];
        if(pedMi[pos*8+2]==tmpPedMi[0*8+2] &&
           pedMi[pos*8+3]==tmpPedMi[0*8+3]) 
        { 
	  for(k=0; k<8; k++)
	  { tmpPedMi[tmpNchild*8+k]=pedMi[pos*8+k]; }
          tmpNchild++; 
	  childVec[j]=-1;
        }
      }
    }
    // get parents info for this nuclear family
    fid=tmpPedMi[0*8+2];
    mid=tmpPedMi[0*8+3];
    tmpNparent=0;
    for(j=0; j<nFamilySize; j++)
    { if((fid==pedMi[j*8+1] || mid==pedMi[j*8+1])
        &&(pedMi[j*8+2]==0)&&(pedMi[j*8+3]==0))
      { pos=tmpNchild+tmpNparent;
        for(k=0; k<8; k++)
        { tmpPedMi[pos*8+k]=pedMi[j*8+k]; }
	tmpNparent++;
      } else if((pedMi[j*8+2]!=0||pedMi[j*8+3]!=0)&&(fid==pedMi[j*8+1] || mid==pedMi[j*8+1]))
      {
	if(*quiet==0)
	{ warning("Warning! The family with id=%d is not a nuclear family", pedMi[j*8+1]); 
	}
      }
    }
    tmpNfamily=tmpNchild+tmpNparent;
    tmpPedMi=(int *)gS_realloc(tmpPedMi, 
                              (size_t)(tmpNfamily*8),
                              (size_t) nFamilySize*8,
                              (size_t) sizeof(int));

    // for one nuclear family
    oneNuclearFamily(marker, tmpPedMi, &tmpNfamily, alleles, nAllele, genoMethod, traitMethod, traitOffset, nX, SM, ESM, CovSM, missingFamily, numInfoM, quiet); 
    //free(tmpPedMi); // Automatically performed for 'S_alloc'
    nLeft -= tmpNchild;
  } while(nLeft>0); 
  return;
}


//void getDimOfFile(char **filename, int *nrow, int *ncol, int *nMarker,
void getDimOfFile(char **filename, int *nrow, int *sampleInfoLen, int *nMarker,
		int *errorFlag)
{
  FILE *fp;
  int iFlag, tmpnr, nSampleInfoLen;

  nSampleInfoLen=*sampleInfoLen;

  fp=fopen(filename[0], "r");
  if(fp==NULL) 
  { printf("Can't open file %s!\n", filename[0]); 
    *errorFlag=-1;
    return; 
  }

  // get number of markers
  iFlag=0;
  *nMarker=countWords(fp, iFlag);
  if(*nMarker==-1) {
    warning("Warning! The file %s does not contain ped data!", filename[0]);
    *errorFlag=-1;
    return;
  }

  // get number of rows
  tmpnr=countLines(fp, *nMarker, nSampleInfoLen);
  if(tmpnr<0) // error occurs
  { *errorFlag=-1;
    return; 
  }

  *nrow=tmpnr;
  fclose(fp);

  return;  
}

int isWhitespace(char c)
{ char whitespace=' ';
  if(c==whitespace)
  { return(1); }
  return(0);
}

// iFlag =1 means we want to check if input is digit
// iFlag =0 means we do not check if input is digit
//int countWords(FILE *fp, fpos_t pos, int iFlag)
int countWords(FILE *fp, int iFlag)
{
  int nc;
  char c, endOfLine='\n', currWhitespace, prevWhitespace=1;

  /**
  flag=fsetpos(fp, &pos);
  if(flag!=0)
  { error("Error in fsetpos()!");
  }
  **/

  // get column numbers
  nc = 0;
  do { 
    c=fgetc(fp);
    if(feof(fp)) // reach end of file
    { nc=-1; break; }
    if( c == endOfLine )
    { break;}

    currWhitespace=isWhitespace(c);
    if((iFlag!=0)&&(currWhitespace==0)&&(isalnum(c))&&(!isdigit(c))) //not a digit
    { printf("c=%d, nc=%d\n", c, nc);
      return(-(nc+2)); 
    }

    
    if( currWhitespace==0 && (isalnum(c)) && prevWhitespace==1)
    { nc++; }
    prevWhitespace=currWhitespace;


  } while(1);
  return(nc);
}

int countLines(FILE *fp, int nMarker, int nSampleInfoLen)
{
  int nr, tmpnc, tmpnc2, tmpnc3, flag, iFlag;
  fpos_t pos;

  nr = 0;
  tmpnc2 = 0;
  tmpnc3=nMarker*2+nSampleInfoLen;
  while(1)
  {
    flag=fgetpos(fp, &pos);
    if(flag!=0)
    { error("Error in fgetpos()!");
    }
    // check if input is digit or not
    iFlag=1;
    tmpnc=countWords(fp, iFlag);
    if(tmpnc==0)
    { break; }
    if(tmpnc==-1) { break; }
    if(tmpnc<0) // error occurs
    { printf("Integer expected. (Line=%d, column=%d) is not an integer!\n",
        (nr+2), -(tmpnc+2)+1);
      return(tmpnc);
    }
    if(nr>0)
    { if(tmpnc!=tmpnc2)
      { printf("number of fields inconsistent with number of markers!\n");
        printf("nMarker=%d, number of fields in line %d is %d\n", nMarker,
                        nr+1, tmpnc);
        printf("expected number of fields in line %d is %d\n",
                        nr+1, tmpnc3);
        return(-1);
      }
    } else
    { tmpnc2=tmpnc; }
    if(tmpnc!=tmpnc3)
    { printf("number of fields inconsistent with number of markers!\n");
      printf("nMarker=%d, number of fields in line %d is %d\n", nMarker, 
		      nr+1, tmpnc);
        printf("expected number of fields in line %d is %d\n",
                        nr+1, tmpnc3);
      return(-1);
    }
    ++nr;
  }
  return(nr);
}

void getPed(char **filename, int *ped, int *nrow, int *ncol, int *nMarker, 
		char **markerNames, int *errorFlag)
{
  int nr, nc, nM, i, j;
  FILE *fp;
  char *Names;

  nr=*nrow;
  nc=*ncol;
  nM=*nMarker;

  Names=(char *)gS_alloc((size_t) nM*50, sizeof(char));

  fp=fopen(filename[0], "r");
  if(fp==NULL)
  { printf("cannot open the file %s!n", filename[0]);
    *errorFlag=-1;
    return;
  }

 //read marker names
  for(i=0; i<nM; i++)
  {
    fscanf(fp, "%s", markerNames[i]);
  }
  fscanf(fp, "\n");


  for(i=0; i<nr; i++)
  { for(j=0; j<nc; j++)
    { fscanf(fp, "%d", &ped[i*nc+j]); 
    }
    fscanf(fp, "\n");
  }    
		 
  fclose(fp);
  //free(Names); // Automatically performed for 'S_alloc'
  *errorFlag=0;
  return;
}

// get observed genotype counts for testing Hardy-Weinberg Equilibrium
void getOij(int *mat, int *nr, int *nc, int *alleles, int *nAllele,
		int *genotypeVec, int *genotypeFreq, int *nGenotype, 
		int *nHET, int *nHOM,
		int *OijMat, double *EijMat, double *piVec, double *chi2)
{
  int i, j, k, m, nnAllele, nnr, nnc, nnGenotype, Oij;
  double Eij;

  nnAllele=*nAllele;
  nnr=*nr;
  nnc=*nc;
  nnGenotype=*nGenotype;

  // get genotypes 
  k=0;
  for(i=0; i<nnAllele; i++)
  { for(j=i; j<nnAllele; j++)
    {
      genotypeVec[k*2+0]=alleles[i];
      genotypeVec[k*2+1]=alleles[j];
      k++; 
    }
  }

  //get observed genotype counts
  //OijMat[i,j]=# of genotype AiAj or AjAi
  //OijMat[i,j]=OijMat[j,i]
  *nHET=0;
  *nHOM=0;
  for(k=0; k<nnr; k++)
  { for(i=0; i<nnAllele; i++)
    { for(j=0; j<=i; j++)
      { if( (mat[k*nnc+0]==alleles[i]&&mat[k*nnc+1]==alleles[j]) ||
	    (mat[k*nnc+0]==alleles[j]&&mat[k*nnc+1]==alleles[i]) )
	{ OijMat[i*nnAllele+j]++; 
	  OijMat[j*nnAllele+i]=OijMat[i*nnAllele+j]; 
	}
      }
    }
    if(mat[k*nnc+0]!=mat[k*nnc+1])
    { (*nHET)++; } // number of heterozygous
    else { (*nHOM)++; } // number of homozygous
  }

  //get observed genotype counts
  for(i=0; i<nnAllele; i++)
  { for(j=i; j<nnAllele; j++)
    { m=nnAllele*((i+1)-1)-((i+1)-1)*((i+1)-2)/2+((j+1)-(i+1));
      genotypeFreq[m]=OijMat[i*nnAllele+j];
    }
  }
  
  //get allele counting
  for(i=0; i<nnAllele; i++)
  { piVec[i]=2.0*OijMat[i*nnAllele+i];
    for(j=0; j<nnAllele; j++)
    { if(j!=i)
      { piVec[i]+=OijMat[i*nnAllele+j]; }
    }
    piVec[i]/=2.0*nnr;
  }

  //get expected genotype counts
  for(i=0; i<nnAllele; i++)
  { EijMat[i*nnAllele+i]=nnr*pow(piVec[i], 2);
    for(j=0; j<i; j++)
    { EijMat[i*nnAllele+j]=2.0*nnr*piVec[i]*piVec[j]; 
      EijMat[j*nnAllele+i]=EijMat[i*nnAllele+j];
    }
  }

 //get chisquare
  *chi2=0;
  for(i=0; i<nnAllele; i++)
  { for(j=0; j<=i; j++)
    { Oij=OijMat[i*nnAllele+j];
      Eij=EijMat[i*nnAllele+j];
      *chi2 += pow(Oij-Eij, 2.0)/Eij;
    }
  }
  
  return;
}

// count genotypes
void getGFreq(int *mat, int *nr, int *nMarker, int *gMat, int *nGenotype, 
		int *countMat)
{
  int i, k, m, start, end, nnMarker, nnGenotype, nnr, nnc;

  nnMarker=*nMarker; // number of alleles
  nnr=*nr; // number of individuals
  nnc=2*nnMarker; // 2*nnMarker 
  nnGenotype=*nGenotype; //number of possible genotypes

  //count numbers of homo/heterozygotes and flag homo/heterozygotes
  //1 -- homo; 0 -- heter; -1 -- genotype contains one missing allele
  //-2 -- genotype contains two missing alleles
  for(m=0; m<nnMarker; m++)
  { 
    start=2*m;
    end=2*m+1;
    for(i=0; i<nnr; i++)
    { //count the genotypes
      for(k=0; k<nnGenotype; k++)
      {
        if((mat[i*nnc+start]==gMat[k*2+0]&& mat[i*nnc+end]==gMat[k*2+1]) || 
           (mat[i*nnc+start]==gMat[k*2+1]&& mat[i*nnc+end]==gMat[k*2+0]))
	{
          (countMat[m*nnGenotype+k])++;
	}
      }
    }
  }

  return;
}

// flag homo/heterozygotes
void FlagHomo(int *mat, int *nr, int *nMarker, int *nHET, int *nHOM, 
              int *nMiss1, int*nMiss2, int *flagHomo)
{
  int i, m, start, end, nnMarker, nnr, nnc;

  nnMarker=*nMarker; // number of alleles
  nnr=*nr; // number of individuals
  nnc=2*nnMarker; // 2*nnMarker 

  //count numbers of homo/heterozygotes and flag homo/heterozygotes
  //1 -- homo; 0 -- heter; -1 -- genotype contains one missing allele
  //-2 -- genotype contains two missing alleles
  for(m=0; m<nnMarker; m++)
  { nHET[m]=0;
    nHOM[m]=0;
    start=2*m;
    end=2*m+1;
    for(i=0; i<nnr; i++)
    { if(mat[i*nnc+start]!=0 && mat[i*nnc+end]!=0)
      { if( mat[i*nnc+start]!=mat[i*nnc+end] )
        { flagHomo[i*nnMarker+m]=0; 
          (nHET[m])++;
        }
        else { 
          flagHomo[i*nnMarker+m]=1; 
          (nHOM[m])++;
        }
      } else { // the genotype contains at least one missing allele
	if(mat[i*nnc+start]+mat[i*nnc+end]>0)
	{ flagHomo[i*nnMarker+m]= -1;  // one missing allele
          (nMiss1[m])++;
        }
	else
	{ flagHomo[i*nnMarker+m]= -2;   // two missing alleles
	  (nMiss2[m])++;
	}
      }
    }
  }

  return;
}


// get observed allele frequencies
void getAFreq(int *mat, int *nr, int *nMarker, int *alleles, int *nAllele, 
		int *aFreqMat)
{
  int i, j, m, start, end, nnMarker, nnAllele, nnr, nnc;

  nnMarker=*nMarker; // number of alleles
  nnr=*nr; // number of individuals
  nnc=2*nnMarker; // 2*nnMarker 
  nnAllele=*nAllele; // number of unique alleles

  //count numbers of homo/heterozygotes and flag homo/heterozygotes
  //1 -- homo; 0 -- heter; -1 -- genotype contains missing alleles
  for(m=0; m<nnMarker; m++)
  { 
    start=2*m;
    end=2*m+1;
    for(i=0; i<nnr; i++)
    { for(j=0; j<nnAllele; j++)
      { if(mat[i*nnc+start]==alleles[j])
	{ (aFreqMat[m*nnAllele+j])++; } 
        if(mat[i*nnc+end]==alleles[j])
	{ (aFreqMat[m*nnAllele+j])++; } 
      }
    }
  }

  return;
}

void getAllGenotypes(int *alleles, int *nAllele, int *Mat, 
		     char **genotypeNames)
{
  int i, j, k, nnAllele;

  nnAllele=*nAllele;
  k=0;
  for(i=0; i<nnAllele; i++)
  { for(j=i; j<nnAllele; j++) 
    {
      Mat[k*2+0]=alleles[i];
      Mat[k*2+1]=alleles[j];
      sprintf(genotypeNames[k], "%d/%d", alleles[i], alleles[j]);
      k++;
    }
  }
  return;
}


void myPrintMatD(double *config, int nrow, int ncol)
{
  int i, j;

  for(i=0; i<nrow; i++)
  { for(j=0; j<ncol; j++)
    { printf("%lf ", config[i*ncol+j]); }
    printf("\n");
  }
  return;
}

void myPrintMatI(int *config, int nrow, int ncol)
{
  int i, j;

  for(i=0; i<nrow; i++)
  { for(j=0; j<ncol; j++)
    { printf("%d ", config[i*ncol+j]); }
    printf("\n");
  }
  return;
}

/* A graceful version of S_alloc() */
void *gS_alloc(size_t n, size_t sizeof_something)
{
  void *p;

  if((p = S_alloc((long) n, (int) sizeof_something)) == NULL){
    error("\nS_alloc() failed - bye!");
  }
  return(p);
}

/* A graceful version of gS_realloc() */
void *gS_realloc(void *p, size_t new, size_t old, size_t sizeof_something)
{
  void *pnew;

  if((pnew = S_realloc(p, (long) new, (long) old, (int) sizeof_something)) == NULL){
    error("\ngS_realloc() failed - bye!");
  }
  return(pnew);
}



/* Transpose of a n by p matrix A */
/* tA = A^t -- p by n                */
void matTranspose(double *A, int *nrow, int *ncol, double *tA)
{
  int i , j, n, p;
  
  n = *nrow; 
  p = *ncol;

  for(i = 0; i < p; i++)    
  { for(j = 0; j < n; j++)
    { tA[i*n+j] = A[j*p+i]; }
  }
  return;
}

/* calculate A = a * A, where a is a scalor and A is a matrix or vector */
void matScalar(double *A, double *a, int *nrow, int *ncol)
{
  int i, j, n, p;
  double aa;

  n=*nrow;
  p=*ncol;
  aa=*a;

  for(i = 0; i < n; i ++){
    for(j = 0; j < p; j ++){
      A[i*p+j] *= aa;
    }
  }
  return;
}

/* Sum of two matrices   */
/* x -- n by p           
 * y -- n by p           
 * out=x+y -- n by p 	 
*/

void matAdd(double *x, int *n, int *p, double *y, double *out)
{
 int nn, pp, i, j;

 nn = *n;
 pp = *p;

 for(i = 0; i < nn; i++)
 { for(j =0; j < pp; j++)
   { out[i*pp+j] = x[i*pp+j] + y[i*pp+j]; }
 }  
 return;
}

/* Substract of two matrices   */
/* x -- n by p           
 * y -- n by p           
 * out=x-y -- n by p 	 
*/

void matSub(double *x, int *n, int *p, double *y, double *out)
{
 int nn, pp, i, j;

 nn = *n;
 pp = *p;

 for(i = 0; i < nn; i++)
 { for(j =0; j < pp; j++)
   { out[i*pp+j] = x[i*pp+j] - y[i*pp+j]; }
 }  
 return;
}


/*
 * C=C+a*A*B
 */
void sumaAB(double *A, int *nrowA, int *ncolA, double *B, int *ncolB, 
		double *a, double *C)
{
  int n, pA, pB;
  double *tmp1, *tmp2, aa;

  n=*nrowA;
  pA=*ncolA;
  pB=*ncolB;
  aa=*a;

  tmp1=(double *)gS_alloc((size_t)(n*pB), sizeof(double));
  tmp2=(double *)gS_alloc((size_t)(n*pB), sizeof(double));

  //tmp1=A*B
  matProd(A,  nrowA, ncolA, B, ncolB, tmp1);

  // tmp1=a*tmp1
  matScalar(tmp1, a, nrowA, ncolB);
  // tmp2=C+tmp1
  matAdd(C, nrowA, ncolB, tmp1, tmp2);


  // C=tmp2
  matAssign(tmp2, nrowA, ncolB, C);

  //free(tmp1); // Automatically performed for 'S_alloc'
  //free(tmp2);  // Automatically performed for 'S_alloc'

  return;
}

/* initialize matrix or vector A with value a */
void matIni(double *A, double *a, int *nrow, int *ncol)
{
  int i, j, n, p;
  double aa;

  n=*nrow;
  p=*ncol;
  aa=*a;

  for(i = 0; i < n; i ++){
    for(j = 0; j < p; j ++){
      A[i*p+j] = aa;
    }
  }
  return;
}

/* initialize matrix or vector A with value a */
void matIniI(int *A, int *a, int *nrow, int *ncol)
{
  int i, j, n, p;
  double aa;

  n=*nrow;
  p=*ncol;
  aa=*a;

  for(i = 0; i < n; i ++){
    for(j = 0; j < p; j ++){
      A[i*p+j] = aa;
    }
  }
  return;
}

/* calculate C = A * B, where A is a nrowA by pA matrix and  B is a pA by pB
 * matrix. C is a nrowA by pB matrix 
 * */
void matProd(double *A, int *nrow, int *pA, double *B, int *pB, double *C)
{
  int i, j, k, n, p1, p2;
  double res;

  n=*nrow;
  p1=*pA;
  p2=*pB;

  for(i = 0; i < n; i ++){
    for(j = 0; j <p2; j ++){
      res = 0.0;
      for(k = 0; k < p1; k ++){
        res += A[i*p1+k] * B[k*p2+j];
      }
      C[i*p2+j] = res;
     }
  }
  return;
}

/* B=A */
void matAssign(double *A, int *nrow, int *ncol, double *B)
{
  int i, j, n, p;

  n=*nrow;
  p=*ncol;

  for(i=0; i<n; i++)
  { for(j=0; j<p; j++)
    { B[i*p+j]=A[i*p+j]; }
  }
  return;
}


/* B=A */
void matAssignI(int *A, int *nrow, int *ncol, int *B)
{
  int i, j, n, p;

  n=*nrow;
  p=*ncol;

  for(i=0; i<n; i++)
  { for(j=0; j<p; j++)
    { B[i*p+j]=A[i*p+j]; }
  }
  return;
}

